#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime

from oslo_versionedobjects import base
from oslo_versionedobjects import fields as obj_fields

# INTRO: This example shows how a object (a plain-old-python-object) with
# some associated fields can be used, and some of its built-in methods can
# be used to convert that object into a primitive and back again (as well
# as determine simple changes on it.


# Ensure that we always register our object with an object registry,
# so that it can be deserialized from its primitive form.
@base.VersionedObjectRegistry.register
class IOTLightbulb(base.VersionedObject):
    """Simple light bulb class with some data about it."""

    VERSION = '1.0'  # Initial version

    #: Namespace these examples will use.
    OBJ_PROJECT_NAMESPACE = 'versionedobjects.examples'

    #: Required fields this object **must** declare.
    fields = {
        'serial': obj_fields.StringField(),
        'manufactured_on': obj_fields.DateTimeField(),
    }


# Now do some basic operations on a light bulb.
bulb = IOTLightbulb(serial='abc-123', manufactured_on=datetime.now())
print(f"The __str__() output of this new object: {bulb}")
print(f"The 'serial' field of the object: {bulb.serial}")
bulb_prim = bulb.obj_to_primitive()
print(f"Primitive representation of this object: {bulb_prim}")

# Now convert the primitive back to an object (isn't it easy!)
bulb = IOTLightbulb.obj_from_primitive(bulb_prim)

bulb.obj_reset_changes()
print(f"The __str__() output of this new (reconstructed) object: {bulb}")

# Mutating a field and showing what changed.
bulb.serial = 'abc-124'
print(
    "After serial number change, the set of fields that"
    f" have been mutated is: {bulb.obj_what_changed()}"
)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Base classes for our unit tests."""

from collections.abc import Callable
import os
from typing import Any, TypeVar
from unittest import mock

import fixtures
from oslo_config import cfg
from oslo_log.fixture import logging_error
import testtools

from oslo_versionedobjects.tests import obj_fixtures


CONF: cfg.ConfigOpts = cfg.CONF
_F = TypeVar('_F', bound=Callable[..., Any])
_C = TypeVar('_C', bound=type)


class TestingException(Exception):
    pass


def _patch_mock_to_raise_for_invalid_assert_calls() -> None:
    def raise_for_invalid_assert_calls(
        wrapped: Callable[[mock.Mock, str], Any],
    ) -> Callable[[mock.Mock, str], Any]:
        def wrapper(_self: mock.Mock, name: str) -> Any:
            valid_asserts = [
                'assert_called_with',
                'assert_called_once_with',
                'assert_has_calls',
                'assert_any_calls',
            ]

            if name.startswith('assert') and name not in valid_asserts:
                raise AttributeError(
                    f'{name} is not a valid mock assert method'
                )

            return wrapped(_self, name)

        return wrapper

    mock.Mock.__getattr__ = raise_for_invalid_assert_calls(  # type: ignore
        mock.Mock.__getattr__
    )


# NOTE(gibi): needs to be called only once at import time
# to patch the mock lib
_patch_mock_to_raise_for_invalid_assert_calls()


class TestCase(testtools.TestCase):
    """Test case base class for all unit tests."""

    TIMEOUT_SCALING_FACTOR: int = 1

    def setUp(self) -> None:
        """Run before each test method to initialize test environment."""
        super().setUp()
        self.useFixture(
            obj_fixtures.Timeout(
                os.environ.get('OS_TEST_TIMEOUT', 0),
                self.TIMEOUT_SCALING_FACTOR,
            )
        )

        self.useFixture(fixtures.NestedTempfile())
        self.useFixture(fixtures.TempHomeDir())
        self.useFixture(obj_fixtures.TranslationFixture())
        self.useFixture(logging_error.get_logging_handle_error_fixture())
        self.useFixture(obj_fixtures.OutputStreamCapture())
        self.useFixture(obj_fixtures.StandardLogging())
        self.useFixture(obj_fixtures.WarningsFixture())

        self.addCleanup(self._clear_attrs)
        self.useFixture(fixtures.EnvironmentVariable('http_proxy'))

    def _clear_attrs(self) -> None:
        # Delete attributes that don't start with _ so they don't pin
        # memory around unnecessarily for the duration of the test
        # suite
        for key in [k for k in self.__dict__.keys() if k[0] != '_']:
            del self.__dict__[key]



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections.abc import Callable
import os

from mypy import nodes
from mypy import plugin as _plugin
from mypy import types


def _fields_dict_from_body(
    body: list[nodes.Statement],
) -> nodes.DictExpr | None:
    """Return the first ``fields = {...}`` DictExpr found in a class body."""
    for statement in body:
        if (
            isinstance(statement, nodes.AssignmentStmt)
            and isinstance(statement.lvalues[0], nodes.NameExpr)
            and statement.lvalues[0].name == "fields"
            and isinstance(statement.rvalue, nodes.DictExpr)
        ):
            return statement.rvalue
    return None


class OsloVersionedObjectPlugin(_plugin.Plugin):
    """A mypy plugin for Oslo VersionedObjects

    The goal of this plugin to add typing information to o.vos during mypy
    static analysis. So that mypy can detect type errors in codes involving
    o.vos.

    It triggers for every class that is decorated with one of the
    VersionedObjectRegistry decorator that generates the o.vo fields runtime
    (e.g. register, register_if, objectify). Then analyze the `fields`
    declaration in the class body to gather the o.vo fields. Then maps the
    type of the field to python types and insert such typed field definition
    to the class definition.

    The plugin also handles inherited fields (e.g. from TimestampedObject
    mixins) by traversing the MRO and reading each parent class's ``fields``
    dict directly from its body. Parent class bodies have already had semantic
    analysis applied (and so have fully-resolved ``fullname`` attributes on
    their AST nodes) by the time subclasses are processed.
    """

    def get_class_decorator_hook(
        self, fullname: str
    ) -> Callable[[_plugin.ClassDefContext], None] | None:
        dec_classes = os.environ.get(
            "OVO_MYPY_DECORATOR_CLASSES", "VersionedObjectRegistry"
        )
        if any(dec_class in fullname for dec_class in dec_classes.split()):
            return self.generate_ovo_field_defs
        return None

    def get_class_decorator_hook_2(
        self, fullname: str
    ) -> Callable[[_plugin.ClassDefContext], bool] | None:
        dec_classes = os.environ.get(
            "OVO_MYPY_DECORATOR_CLASSES", "VersionedObjectRegistry"
        )
        if any(dec_class in fullname for dec_class in dec_classes.split()):
            return self.generate_ovo_field_defs_2
        return None

    def get_base_class_hook(
        self, fullname: str
    ) -> Callable[[_plugin.ClassDefContext], None] | None:
        base_classes = os.environ.get(
            "OVO_MYPY_BASE_CLASSES", "VersionedObject VersionedObjectMixin"
        )
        if any(base_class in fullname for base_class in base_classes.split()):
            return self.generate_ovo_field_defs
        return None

    def _add_member_to_class(
        self, member_name: str, member_type: types.Type, clazz: nodes.TypeInfo
    ) -> None:
        """Add a new member to the class.

        Add a variable with given name and type to the symbol table of a
        class. This also takes care about setting necessary attributes on the
        variable node.
        """
        var = nodes.Var(member_name)
        var.info = clazz
        var._fullname = clazz.fullname + "." + member_name
        var.type = member_type
        clazz.names[member_name] = nodes.SymbolTableNode(nodes.MDEF, var)
        self.log(
            f"Defined o.vo field: {clazz.fullname}.{member_name} as "
            f"{member_type}"
        )

    def _apply_nullable(
        self,
        field_type: types.Type,
        ctx: _plugin.ClassDefContext,
        kwargs: dict[str, nodes.Expression],
    ) -> types.Type:
        if "nullable" in kwargs and ctx.api.parse_bool(kwargs["nullable"]):
            return types.UnionType([field_type, types.NoneType()])
        return field_type

    def _resolve_ovo_class_type(
        self,
        ctx: _plugin.ClassDefContext,
        class_name: str,
    ) -> types.Type | None:
        """Look up a versioned object class by name and return its mypy Type.

        Tries the current module scope first, then falls back to searching all
        loaded modules.
        """
        sym = ctx.api.lookup_qualified(
            class_name, ctx.cls, suppress_errors=True
        )
        if sym is not None and isinstance(sym.node, nodes.TypeInfo):
            return types.Instance(sym.node, [])

        for module in ctx.api.modules.values():
            node = module.names.get(class_name)
            if node is not None and isinstance(node.node, nodes.TypeInfo):
                return types.Instance(node.node, [])

        return None

    def _get_python_type_from_ovo_field_type(
        self,
        ctx: _plugin.ClassDefContext,
        ovo_field_type_name: str,
        args: list[nodes.Expression],
        kwargs: dict[str, nodes.Expression],
    ) -> types.Type | None:
        # lookup_fully_qualified_or_none requires a dotted name (bare names
        # like a local callable would raise ValueError inside mypy)
        if '.' not in ovo_field_type_name:
            self.log(f"Unqualified field type name: {ovo_field_type_name}")
            return types.AnyType(types.TypeOfAny.implementation_artifact)

        field_symbol = ctx.api.lookup_fully_qualified_or_none(
            ovo_field_type_name
        )
        if field_symbol is None or not isinstance(
            field_symbol.node, nodes.TypeInfo
        ):
            self.log(f"Could not find field type {ovo_field_type_name}")
            return types.AnyType(types.TypeOfAny.implementation_artifact)

        field_fullname = field_symbol.node.fullname

        # ObjectField and ListOfObjectsField take the target class name as a
        # positional string arg rather than a generic parameter
        if field_fullname == 'oslo_versionedobjects.fields.ListOfObjectsField':
            base_type: types.Type | None = None
            if args and isinstance(args[0], nodes.StrExpr):
                resolved = self._resolve_ovo_class_type(ctx, args[0].value)
                if resolved is not None:
                    list_sym = ctx.api.lookup_fully_qualified_or_none(
                        'builtins.list'
                    )
                    if list_sym and isinstance(list_sym.node, nodes.TypeInfo):
                        base_type = types.Instance(list_sym.node, [resolved])
            if base_type is None:
                self.log(
                    f"Could not resolve object type for {ovo_field_type_name}"
                )
                return None
            return self._apply_nullable(base_type, ctx, kwargs)

        if field_fullname == 'oslo_versionedobjects.fields.ObjectField':
            if args and isinstance(args[0], nodes.StrExpr):
                resolved = self._resolve_ovo_class_type(ctx, args[0].value)
                if resolved is not None:
                    return self._apply_nullable(resolved, ctx, kwargs)
            self.log(
                f"Could not resolve object type for {ovo_field_type_name}"
            )
            return None

        # AutoTypedField is a proper generic. We can retrieve its type from
        # this.
        for class_info in field_symbol.node.mro:
            for base in class_info.bases:
                if (
                    isinstance(base, types.Instance)
                    and base.type.fullname
                    == 'oslo_versionedobjects.fields.AutoTypedField'
                    and base.args
                    and not isinstance(base.args[0], types.TypeVarType)
                ):
                    return self._apply_nullable(base.args[0], ctx, kwargs)

        return types.AnyType(types.TypeOfAny.implementation_artifact)

    def _add_ovo_members_to_class(
        self,
        ctx: _plugin.ClassDefContext,
        fields_def: nodes.DictExpr,
        processed_fields: set[str],
    ) -> bool:
        all_resolved = True

        for k, v in fields_def.items:
            # This means we do not support the case when the name of the
            # field is calculated e.g.:
            # fields = {'first' + 'name': fields.StringField()}
            if not isinstance(k, nodes.StrExpr):
                ctx.api.fail(
                    "oslo.versionedobject `fields` dict should have string "
                    "literal keys",
                    ctx.cls,
                )
                continue

            field_name = k.value

            # Skip fields already defined by a more derived class in the MRO
            if field_name in processed_fields:
                continue

            processed_fields.add(field_name)

            if (
                not isinstance(v, nodes.CallExpr)
                or not isinstance(v.callee, (nodes.MemberExpr, nodes.NameExpr))
                or not v.callee.fullname
            ):
                self.log(
                    f"Skipping field {field_name}: unexpected AST structure"
                )
                field_type: types.Type = types.AnyType(
                    types.TypeOfAny.implementation_artifact
                )
            else:
                args = [
                    arg
                    for arg, arg_name in zip(v.args, v.arg_names)
                    if arg_name is None
                ]
                kwargs = {
                    arg_name: arg
                    for arg, arg_name in zip(v.args, v.arg_names)
                    if arg_name is not None
                }

                resolved = self._get_python_type_from_ovo_field_type(
                    ctx, v.callee.fullname, args, kwargs
                )
                if resolved is None:
                    all_resolved = False
                    field_type = types.AnyType(
                        types.TypeOfAny.implementation_artifact
                    )
                else:
                    field_type = resolved

            self._add_member_to_class(field_name, field_type, ctx.cls.info)

        return all_resolved

    def generate_ovo_field_defs(self, ctx: _plugin.ClassDefContext) -> None:
        self.generate_ovo_field_defs_2(ctx)

    def generate_ovo_field_defs_2(self, ctx: _plugin.ClassDefContext) -> bool:
        # Process fields from this class and all classes in its MRO whose
        # bodies are still available (i.e. same-file classes).  Cross-module
        # parent classes have their bodies cleared by mypy after their own
        # module is analyzed; their fields are instead picked up via mypy's
        # normal MRO attribute resolution because the plugin fires for those
        # classes too (via get_base_class_hook) while their bodies are intact.
        #
        # hook_2 callables can return False to request a retry, but we always
        # return True: by the time hook_2 fires all modules are loaded, so any
        # ObjectField target that is still unresolvable will remain so on
        # subsequent attempts.
        processed_fields: set[str] = set()

        for type_info in ctx.cls.info.mro:
            fields_dict_expr = _fields_dict_from_body(type_info.defn.defs.body)
            if fields_dict_expr is None:
                continue

            self._add_ovo_members_to_class(
                ctx, fields_dict_expr, processed_fields
            )

        return True

    def log(self, msg: str) -> None:
        if self.options.verbosity > 0:
            print("LOG:  OsloVersionedObjectPlugin: " + msg)


def plugin(version: str) -> type[_plugin.Plugin]:
    return OsloVersionedObjectPlugin



#    Copyright 2011 Justin Santa Barbara
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

from oslo_versionedobjects import exception
from oslo_versionedobjects import test

notifier = mock.Mock()


class TestWrapper:
    @exception.wrap_exception(notifier=notifier)
    def raise_exc(self, context, exc, admin_password):
        raise exc


class ExceptionTestCase(test.TestCase):
    def test_wrap_exception_wrapped(self):
        test = TestWrapper()
        # Ensure that the original function is available in
        # the __wrapped__ attribute
        self.assertTrue(hasattr(test.raise_exc, '__wrapped__'))

    def test_wrap_exception(self):
        context = "context"
        exc = ValueError()

        test = TestWrapper()
        notifier.reset_mock()

        # wrap_exception() must reraise the exception
        self.assertRaises(
            ValueError,  # nosec
            test.raise_exc,
            context,
            exc,
            admin_password="xxx",
        )

        # wrap_exception() strips admin_password from args
        payload = {
            'args': {'self': test, 'context': context, 'exc': exc},
            'exception': exc,
        }
        notifier.error.assert_called_once_with(context, 'raise_exc', payload)

    def test_vo_exception(self):
        exc = exception.VersionedObjectsException()
        self.assertEqual('An unknown exception occurred.', str(exc))
        self.assertEqual({'code': 500}, exc.kwargs)

    def test_object_action_error(self):
        exc = exception.ObjectActionError(
            action='ACTION', reason='REASON', code=123
        )
        self.assertEqual(
            'Object action ACTION failed because: REASON', str(exc)
        )
        self.assertEqual(
            {'code': 123, 'action': 'ACTION', 'reason': 'REASON'}, exc.kwargs
        )

    def test_constructor_format_error(self):
        # Test error handling on formatting exception message in the
        # VersionedObjectsException constructor
        with mock.patch.object(exception, 'LOG') as log:
            exc = exception.ObjectActionError()

            log.error.assert_called_with('code: 500')

        # Formatting failed: the message is the original format string
        self.assertEqual(exception.ObjectActionError.msg_fmt, str(exc))



#    Copyright 2013 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import abc
from collections.abc import Collection, Iterable, Mapping
import copy
import datetime
import logging
import re
from typing import (
    Any,
    cast,
    Generic,
    overload,
    SupportsIndex,
    Self,
    TypeVar,
    TYPE_CHECKING,
)
import uuid
import warnings

import netaddr
from oslo_utils import strutils
from oslo_utils import timeutils
from oslo_utils import versionutils

from oslo_versionedobjects._i18n import _
from oslo_versionedobjects import _utils
from oslo_versionedobjects import exception

if TYPE_CHECKING:
    from oslo_versionedobjects import base

LOG = logging.getLogger('field')

# Type variable for generic Field type
T = TypeVar('T')


class KeyTypeError(TypeError):
    def __init__(self, expected: type, value: Any) -> None:
        super().__init__(
            _('Key %(key)s must be of type %(expected)s not %(actual)s')
            % {
                'key': repr(value),
                'expected': expected.__name__,
                'actual': value.__class__.__name__,
            }
        )


class ElementTypeError(TypeError):
    def __init__(self, expected: str, key: Any, value: Any) -> None:
        super().__init__(
            _(
                'Element %(key)s:%(val)s must be of type %(expected)s'
                ' not %(actual)s'
            )
            % {
                'key': key,
                'val': repr(value),
                'expected': expected,
                'actual': value.__class__.__name__,
            }
        )


U = TypeVar('U')
E = TypeVar('E')  # Element type for compound fields (List, Dict, Set)


class AbstractFieldType(Generic[T], metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> T:
        """This is called to coerce (if possible) a value on assignment.

        This method should convert the value given into the designated type,
        or throw an exception if this is not possible.

        :param obj: The VersionedObject on which an attribute is being set
        :param attr: The name of the attribute being set
        :param value: The value being set
        :returns: A properly-typed value
        """
        ...

    @abc.abstractmethod
    def from_primitive(
        self, obj: base.VersionedObject, attr: str, value: object
    ) -> T:
        """This is called to deserialize a value.

        This method should deserialize a value from the form given by
        to_primitive() to the designated type.

        :param obj: The VersionedObject on which the value is to be set
        :param attr: The name of the attribute which will hold the value
        :param value: The serialized form of the value
        :returns: The natural form of the value
        """
        ...

    # TODO(stephenfin): We could probably type this better with a second
    # generic parameter, but do we care enough?
    @abc.abstractmethod
    def to_primitive(
        self, obj: base.VersionedObject, attr: str, value: T
    ) -> object:
        """This is called to serialize a value.

        This method should serialize a value to the form expected by
        from_primitive().

        :param obj: The VersionedObject on which the value is set
        :param attr: The name of the attribute holding the value
        :param value: The natural form of the value
        :returns: The serialized form of the value
        """
        ...

    @abc.abstractmethod
    def describe(self) -> str:
        """Returns a string describing the type of the field."""
        ...

    @abc.abstractmethod
    def stringify(self, value: T) -> str:
        """Returns a short stringified version of a value."""
        ...


class FieldType(AbstractFieldType[T]):
    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> T:
        return cast(T, value)

    def from_primitive(
        self, obj: base.VersionedObject, attr: str, value: object
    ) -> T:
        return cast(T, value)

    def to_primitive(
        self, obj: base.VersionedObject, attr: str, value: T
    ) -> object:
        return value

    def describe(self) -> str:
        return self.__class__.__name__

    def stringify(self, value: T) -> str:
        return str(value)

    # TODO(stephenfin): This is a minimal JSON Schema schema. We can type this
    # better.
    def get_schema(self) -> dict[str, Any]:
        raise NotImplementedError()


class UnspecifiedDefault:
    pass


class Field(Generic[T]):
    _type: FieldType[T]
    _nullable: bool
    _default: T | type[UnspecifiedDefault]
    _read_only: bool

    def __init__(
        self,
        field_type: FieldType[T],
        nullable: bool = False,
        default: T | type[UnspecifiedDefault] = UnspecifiedDefault,
        read_only: bool = False,
    ) -> None:
        self._type = field_type
        self._nullable = nullable
        self._default = default
        self._read_only = read_only

    def __repr__(self) -> str:
        if isinstance(self._default, set):
            # TODO(stephenfin): Drop this when we switch from
            # 'inspect.getargspec' to 'inspect.getfullargspec', since our
            # hashes will have to change anyway
            # make a py27 and py35 compatible representation. See bug 1771804
            default = 'set([{}])'.format(
                ','.join(sorted([str(v) for v in self._default]))
            )
        else:
            default = str(self._default)
        return (
            f'{self._type.__class__.__name__}'
            f'(default={default},nullable={self._nullable})'
        )

    @property
    def nullable(self) -> bool:
        return self._nullable

    @property
    def default(self) -> T | type[UnspecifiedDefault]:
        return self._default

    @property
    def read_only(self) -> bool:
        return self._read_only

    def _null(self, obj: base.VersionedObject | None, attr: str) -> T | None:
        if self.nullable:
            return None

        if self._default != UnspecifiedDefault:
            # NOTE(danms): We coerce the default value each time the field
            # is set to None as our contract states that we'll let the type
            # examine the object and attribute name at that time.
            return self._type.coerce(obj, attr, copy.deepcopy(self._default))

        raise ValueError(_("Field `%s' cannot be None") % attr)

    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> T | None:
        """Coerce a value to a suitable type.

        This is called any time you set a value on an object, like:

          foo.myint = 1

        and is responsible for making sure that the value (1 here) is of
        the proper type, or can be sanely converted.

        This also handles the potentially nullable or defaultable
        nature of the field and calls the coerce() method on a
        FieldType to actually do the coercion.

        :param obj: The object being acted upon
        :param attr: The name of the attribute/field being set
        :param value: The value being set
        :returns: The properly-typed value
        """
        if value is None:
            # _null may return None (if nullable) which is valid for T when
            # T includes None; otherwise it raises or returns default
            return self._null(obj, attr)

        return self._type.coerce(obj, attr, value)

    @overload
    def from_primitive(
        self, obj: base.VersionedObject, attr: str, value: None
    ) -> None: ...

    @overload
    def from_primitive(
        self, obj: base.VersionedObject, attr: str, value: object
    ) -> T: ...

    def from_primitive(
        self, obj: base.VersionedObject, attr: str, value: object | None
    ) -> T | None:
        """Deserialize a value from primitive form.

        This is responsible for deserializing a value from primitive
        into regular form. It calls the from_primitive() method on a
        FieldType to do the actual deserialization.

        :param obj: The object being acted upon
        :param attr: The name of the attribute/field being deserialized
        :param value: The value to be deserialized
        :returns: The deserialized value
        """
        if value is None:
            return None

        return self._type.from_primitive(obj, attr, value)

    def to_primitive(
        self, obj: base.VersionedObject, attr: str, value: T
    ) -> object:
        """Serialize a value to primitive form.

        This is responsible for serializing a value to primitive
        form. It calls to_primitive() on a FieldType to do the actual
        serialization.

        :param obj: The object being acted upon
        :param attr: The name of the attribute/field being serialized
        :param value: The value to be serialized
        :returns: The serialized value
        """
        if value is None:
            return None

        return self._type.to_primitive(obj, attr, value)

    def describe(self) -> str:
        """Return a short string describing the type of this field."""
        name = self._type.describe()
        prefix = self.nullable and 'Nullable' or ''
        return prefix + name

    def stringify(self, value: T) -> str:
        if value is None:
            return 'None'

        return self._type.stringify(value)

    # TODO(stephenfin): This is a minimal JSON Schema schema. We can type this
    # better.
    def get_schema(self) -> dict[str, Any]:
        schema = self._type.get_schema()
        schema.update({'readonly': self.read_only})
        if self.nullable:
            schema['type'].append('null')
        default = self.default
        if default != UnspecifiedDefault:
            schema.update({'default': default})
        return schema


class String(FieldType[str]):
    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> str:
        # FIXME(danms): We should really try to avoid the need to do this
        accepted_types = (int, float, str, datetime.datetime)
        if isinstance(value, accepted_types):
            return str(value)

        raise ValueError(
            _('A string is required in field %(attr)s, not a %(type)s')
            % {'attr': attr, 'type': type(value).__name__}
        )

    def stringify(self, value: str) -> str:
        return f'\'{value}\''

    def get_schema(self) -> dict[str, Any]:
        return {'type': ['string']}


class SensitiveString(String):
    """A string field type that may contain sensitive (password) information.

    Passwords in the string value are masked when stringified.
    """

    def stringify(self, value: str) -> str:
        return super().stringify(strutils.mask_password(value))


class VersionPredicate(String):
    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> str:
        if not isinstance(value, str):
            raise ValueError(
                _('Version %(val)s should be a string type, not %(real_type)s')
                % {'val': value, 'real_type': type(value)}
            )
        try:
            versionutils.VersionPredicate(value)
        except ValueError:
            raise ValueError(
                _('Version %(val)s is not a valid predicate in field %(attr)s')
                % {'val': value, 'attr': attr}
            )
        return value


class Enum(String):
    def __init__(self, valid_values: Collection[str]) -> None:
        if not valid_values:
            raise exception.EnumRequiresValidValuesError()

        try:
            # Test validity of the values
            for value in valid_values:
                super().coerce(None, 'init', value)
        except (TypeError, ValueError):
            raise exception.EnumValidValuesInvalidError()

        self._valid_values = valid_values

        super().__init__()

    @property
    def valid_values(self) -> Collection[str]:
        return copy.copy(self._valid_values)

    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> str:
        if value not in self._valid_values:
            msg = _("Field value %s is invalid") % value
            raise ValueError(msg)
        return super().coerce(obj, attr, value)

    def stringify(self, value: str) -> str:
        if value not in self._valid_values:
            msg = _("Field value %s is invalid") % value
            raise ValueError(msg)
        return super().stringify(value)

    def get_schema(self) -> dict[str, Any]:
        schema = super().get_schema()
        schema['enum'] = self._valid_values
        return schema


class StringPattern(FieldType[T]):
    PATTERN: str

    def get_schema(self) -> dict[str, Any]:
        if not hasattr(self, "PATTERN"):
            msg = _("%s has no pattern") % self.__class__.__name__
            raise AttributeError(msg)

        return {'type': ['string'], 'pattern': self.PATTERN}


class UUID(StringPattern[str]):
    PATTERN = (
        r'^[a-fA-F0-9]{8}-?[a-fA-F0-9]{4}-?[a-fA-F0-9]{4}-?[a-fA-F0-9]'
        r'{4}-?[a-fA-F0-9]{12}$'
    )

    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> str:
        # FIXME(danms): We should actually verify the UUIDness here
        with warnings.catch_warnings():
            # Change the warning action only if no other filter exists
            # for this warning to allow the client to define other action
            # like 'error' for this warning.
            warnings.filterwarnings(action="once", append=True)
            try:
                uuid.UUID(f"{value}")
            except Exception:
                # This is to ensure no breaking behaviour for current
                # users
                warnings.warn(
                    f"{value!r} is an invalid UUID. Using UUIDFields "
                    f"with invalid UUIDs is no longer "
                    f"supported, and will be removed in a future "
                    f"release. Please update your "
                    f"code to input valid UUIDs or accept "
                    f"ValueErrors for invalid UUIDs. See "
                    f"https://docs.openstack.org/oslo.versionedobjects/latest/reference/fields.html#oslo_versionedobjects.fields.UUIDField "  # noqa
                    f"for further details",
                    FutureWarning,
                )

            return f"{value}"


class MACAddress(StringPattern[str]):
    PATTERN = r'^[0-9a-f]{2}(:[0-9a-f]{2}){5}$'
    _REGEX = re.compile(PATTERN)

    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> str:
        if isinstance(value, str):
            lowered = value.lower().replace('-', ':')
            if MACAddress._REGEX.match(lowered):
                return lowered
        raise ValueError(_("Malformed MAC %s") % (value,))


class PCIAddress(StringPattern[str]):
    PATTERN = r'^[0-9a-f]{4}:[0-9a-f]{2}:[0-1][0-9a-f].[0-7]$'
    _REGEX = re.compile(PATTERN)

    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> str:
        if isinstance(value, str):
            newvalue = value.lower()
            if PCIAddress._REGEX.match(newvalue):
                return newvalue
        raise ValueError(_("Malformed PCI address %s") % (value,))


class Integer(FieldType[int]):
    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> int:
        return int(value)

    def get_schema(self) -> dict[str, Any]:
        return {'type': ['integer']}


class NonNegativeInteger(FieldType[int]):
    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> int:
        v = int(value)
        if v < 0:
            raise ValueError(_('Value must be >= 0 for field %s') % attr)
        return v

    def get_schema(self) -> dict[str, Any]:
        return {'type': ['integer'], 'minimum': 0}


class Float(FieldType[float]):
    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> float:
        return float(value)

    def get_schema(self) -> dict[str, Any]:
        return {'type': ['number']}


class NonNegativeFloat(FieldType[float]):
    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> float:
        v = float(value)
        if v < 0:
            raise ValueError(_('Value must be >= 0 for field %s') % attr)
        return v

    def get_schema(self) -> dict[str, Any]:
        return {'type': ['number'], 'minimum': 0}


class Boolean(FieldType[bool]):
    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> bool:
        return bool(value)

    def get_schema(self) -> dict[str, Any]:
        return {'type': ['boolean']}


class FlexibleBoolean(Boolean):
    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> bool:
        return strutils.bool_from_string(value)


class DateTime(FieldType[datetime.datetime]):
    def __init__(
        self, tzinfo_aware: bool = True, *args: Any, **kwargs: Any
    ) -> None:
        self.tzinfo_aware = tzinfo_aware
        super().__init__(*args, **kwargs)

    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> datetime.datetime:
        if isinstance(value, str):
            # NOTE(danms): Being tolerant of isotime strings here will help us
            # during our objects transition
            ts = timeutils.parse_isotime(value)
        elif isinstance(value, datetime.datetime):
            ts = value
        else:
            raise ValueError(
                _(
                    'A datetime.datetime is required '
                    'in field %(attr)s, not a %(type)s'
                )
                % {'attr': attr, 'type': type(value).__name__}
            )

        if ts.utcoffset() is None and self.tzinfo_aware:
            # NOTE(danms): Legacy objects from sqlalchemy are stored in UTC,
            # but are returned without a timezone attached.
            # As a transitional aid, assume a tz-naive object is in UTC.
            ts = ts.replace(tzinfo=datetime.UTC)
        elif not self.tzinfo_aware:
            ts = ts.replace(tzinfo=None)
        return ts

    def from_primitive(
        self, obj: base.VersionedObject, attr: str, value: Any
    ) -> datetime.datetime:
        return self.coerce(obj, attr, timeutils.parse_isotime(value))

    def get_schema(self) -> dict[str, Any]:
        return {'type': ['string'], 'format': 'date-time'}

    def to_primitive(
        self, obj: base.VersionedObject, attr: str, value: datetime.datetime
    ) -> object:
        return _utils.isotime(value)

    def stringify(self, value: datetime.datetime) -> str:
        return _utils.isotime(value)


class IPAddress(StringPattern[netaddr.IPAddress]):
    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> netaddr.IPAddress:
        try:
            return netaddr.IPAddress(value)
        except netaddr.AddrFormatError as e:
            raise ValueError(str(e))

    def from_primitive(
        self, obj: base.VersionedObject, attr: str, value: object
    ) -> netaddr.IPAddress:
        return self.coerce(obj, attr, value)

    def to_primitive(
        self, obj: base.VersionedObject, attr: str, value: T
    ) -> object:
        return str(value)


class IPV4Address(IPAddress):
    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> netaddr.IPAddress:
        result = super().coerce(obj, attr, value)
        if result.version != 4:
            raise ValueError(
                _('Network "%(val)s" is not valid in field %(attr)s')
                % {'val': value, 'attr': attr}
            )
        return result

    def get_schema(self) -> dict[str, Any]:
        return {'type': ['string'], 'format': 'ipv4'}


class IPV6Address(IPAddress):
    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> netaddr.IPAddress:
        result = super().coerce(obj, attr, value)
        if result.version != 6:
            raise ValueError(
                _('Network "%(val)s" is not valid in field %(attr)s')
                % {'val': value, 'attr': attr}
            )
        return result

    def get_schema(self) -> dict[str, Any]:
        return {'type': ['string'], 'format': 'ipv6'}


class IPV4AndV6Address(IPAddress):
    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> netaddr.IPAddress:
        result = super().coerce(obj, attr, value)
        if result.version != 4 and result.version != 6:
            raise ValueError(
                _('Network "%(val)s" is not valid in field %(attr)s')
                % {'val': value, 'attr': attr}
            )
        return result

    def get_schema(self) -> dict[str, Any]:
        return {
            'oneOf': [IPV4Address().get_schema(), IPV6Address().get_schema()]
        }


class IPNetwork(StringPattern[netaddr.IPNetwork]):
    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> netaddr.IPNetwork:
        try:
            return netaddr.IPNetwork(value)
        except netaddr.AddrFormatError as e:
            raise ValueError(str(e))

    def from_primitive(
        self, obj: base.VersionedObject, attr: str, value: object
    ) -> netaddr.IPNetwork:
        return self.coerce(obj, attr, value)

    def to_primitive(
        self, obj: base.VersionedObject, attr: str, value: T
    ) -> object:
        return str(value)


class IPV4Network(IPNetwork):
    PATTERN = (
        r'^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-'
        r'9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])(\/([0-9]|[1-2]['
        r'0-9]|3[0-2]))$'
    )

    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> netaddr.IPNetwork:
        try:
            return netaddr.IPNetwork(value, version=4)
        except netaddr.AddrFormatError as e:
            raise ValueError(str(e))


class IPV6Network(IPNetwork):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.PATTERN = self._create_pattern()

    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> netaddr.IPNetwork:
        try:
            return netaddr.IPNetwork(value, version=6)
        except netaddr.AddrFormatError as e:
            raise ValueError(str(e))

    def _create_pattern(self) -> str:
        ipv6seg = '[0-9a-fA-F]{1,4}'
        ipv4seg = '(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])'

        return (
            # Pattern based on answer to
            # http://stackoverflow.com/questions/53497/regular-expression-that-matches-valid-ipv6-addresses
            '^'
            # 1:2:3:4:5:6:7:8
            '(' + ipv6seg + ':){7,7}' + ipv6seg + '|'
            # 1:: 1:2:3:4:5:6:7::
            '(' + ipv6seg + ':){1,7}:|'
            # 1::8 1:2:3:4:5:6::8 1:2:3:4:5:6::8
            '(' + ipv6seg + ':){1,6}:' + ipv6seg + '|'
            # 1::7:8 1:2:3:4:5::7:8 1:2:3:4:5::8
            '(' + ipv6seg + ':){1,5}(:' + ipv6seg + '){1,2}|'
            # 1::6:7:8 1:2:3:4::6:7:8 1:2:3:4::8
            '(' + ipv6seg + ':){1,4}(:' + ipv6seg + '){1,3}|'
            # 1::5:6:7:8 1:2:3::5:6:7:8 1:2:3::8
            '(' + ipv6seg + ':){1,3}(:' + ipv6seg + '){1,4}|'
            # 1::4:5:6:7:8 1:2::4:5:6:7:8 1:2::8
            '('
            + ipv6seg
            + ':){1,2}(:'
            + ipv6seg
            + '){1,5}|'
            +
            # 1::3:4:5:6:7:8 1::3:4:5:6:7:8 1::8
            ipv6seg
            + ':((:'
            + ipv6seg
            + '){1,6})|'
            # ::2:3:4:5:6:7:8 ::2:3:4:5:6:7:8 ::8 ::
            ':((:' + ipv6seg + '){1,7}|:)|'
            # fe80::7:8%eth0 fe80::7:8%1
            'fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|'
            # ::255.255.255.255 ::ffff:255.255.255.255 ::ffff:0:255.255.255.255
            '::(ffff(:0{1,4}){0,1}:){0,1}'
            '(' + ipv4seg + r'\.){3,3}' + ipv4seg + '|'
            # 2001:db8:3:4::192.0.2.33 64:ff9b::192.0.2.33
            '(' + ipv6seg + ':){1,4}:'
            '('
            + ipv4seg
            + r'\.){3,3}'
            + ipv4seg
            +
            # /128
            r'(\/(d|dd|1[0-1]d|12[0-8]))$'
        )


class CoercedCollectionMixin(Generic[T]):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._element_type: Field[T] | None = None
        self._obj: Any = None
        self._field: str | None = None
        super().__init__(*args, **kwargs)

    def enable_coercing(
        self, element_type: Field[T], obj: Any, field: str
    ) -> None:
        self._element_type = element_type
        self._obj = obj
        self._field = field


class CoercedList(CoercedCollectionMixin[T], list[T]):
    """List which coerces its elements

    List implementation which overrides all element-adding methods and
    coercing the element(s) being added to the required element type
    """

    def _coerce_item(self, index: int, item: Any) -> Any:
        if hasattr(self, "_element_type") and self._element_type is not None:
            att_name = f"{self._field}[{index}]"
            return self._element_type.coerce(self._obj, att_name, item)

        return item

    def __setitem__(self, i: int | slice, y: Any) -> None:  # type: ignore[override]
        if type(i) is slice:  # compatibility with py3 and [::] slices
            start = i.start or 0
            step = i.step or 1
            coerced_items = [
                self._coerce_item(start + index * step, item)
                for index, item in enumerate(y)
            ]
            super().__setitem__(i, coerced_items)
        else:
            super().__setitem__(i, self._coerce_item(i, y))

    def append(self, x: Any) -> None:
        super().append(self._coerce_item(len(self) + 1, x))

    def extend(self, t: Any) -> None:
        coerced_items = [
            self._coerce_item(len(self) + index, item)
            for index, item in enumerate(t)
        ]
        super().extend(coerced_items)

    def insert(self, i: SupportsIndex, x: Any) -> None:
        # Convert SupportsIndex to int for _coerce_item which expects int
        index = int(i)
        super().insert(i, self._coerce_item(index, x))

    def __add__(  # type: ignore[override]
        self, y: Iterable[Any]
    ) -> Self:
        result = type(self)(self)
        if self._element_type is not None and self._field is not None:
            result.enable_coercing(self._element_type, self._obj, self._field)
        coerced_items = [
            result._coerce_item(len(result) + index, item)
            for index, item in enumerate(y)
        ]
        result.extend(coerced_items)
        return result

    def __iadd__(  # type: ignore[override]
        self, y: Iterable[Any]
    ) -> Self:
        coerced_items = [
            self._coerce_item(len(self) + index, item)
            for index, item in enumerate(y)
        ]
        super().__iadd__(coerced_items)
        return self

    def __setslice__(self, i: int, j: int, y: Any) -> None:
        coerced_items = [
            self._coerce_item(i + index, item) for index, item in enumerate(y)
        ]
        super().__setslice__(i, j, coerced_items)  # type: ignore[misc]


class CoercedDict(CoercedCollectionMixin[T], dict[str, T]):
    """Dict which coerces its values

    Dict implementation which overrides all element-adding methods and
    coercing the element(s) being added to the required element type
    """

    def _coerce_dict(self, d: dict[str, Any]) -> dict[str, Any]:
        res: dict[str, Any] = {}
        for key, element in d.items():
            res[key] = self._coerce_item(key, element)
        return res

    def _coerce_item(self, key: str, item: Any) -> Any:
        if not isinstance(key, str):
            raise KeyTypeError(str, key)

        if hasattr(self, "_element_type") and self._element_type is not None:
            att_name = f"{self._field}[{key}]"
            return self._element_type.coerce(self._obj, att_name, item)

        return item

    def __setitem__(self, key: str, value: Any) -> None:
        super().__setitem__(key, self._coerce_item(key, value))

    def update(  # type: ignore[override]
        self, other: dict[str, Any] | None = None, **kwargs: Any
    ) -> None:
        if other is not None:
            super().update(
                self._coerce_dict(other), **self._coerce_dict(kwargs)
            )
        else:
            super().update(**self._coerce_dict(kwargs))

    def setdefault(self, key: str, default: Any = None) -> Any:
        return super().setdefault(key, self._coerce_item(key, default))


class CoercedSet(CoercedCollectionMixin[T], set[T]):
    """Set which coerces its values

    Set implementation which overrides all element-adding methods and
    coercing the element(s) being added to the required element type
    """

    def _coerce_element(self, element: Any) -> Any:
        if hasattr(self, "_element_type") and self._element_type is not None:
            return self._element_type.coerce(
                self._obj, f"{self._field}[{element}]", element
            )

        return element

    def _coerce_iterable(self, values: Iterable[Any]) -> set[T]:
        coerced = set()
        for element in values:
            coerced.add(self._coerce_element(element))
        return coerced

    def add(self, value: Any) -> None:
        super().add(self._coerce_element(value))

    def update(self, values: Iterable[Any]) -> None:  # type: ignore[override]
        super().update(self._coerce_iterable(values))

    def symmetric_difference_update(self, values: Iterable[Any]) -> None:
        super().symmetric_difference_update(self._coerce_iterable(values))

    def __or__(  # type: ignore[override]
        self, y: Iterable[Any]
    ) -> Self:
        result = type(self)(self)
        if self._element_type is not None and self._field is not None:
            result.enable_coercing(self._element_type, self._obj, self._field)
        result.update(y)
        return result

    def __ior__(  # type: ignore[override]
        self, y: Iterable[Any]
    ) -> Self:
        return super().__ior__(self._coerce_iterable(y))

    def __xor__(  # type: ignore[override]
        self, y: Iterable[Any]
    ) -> Self:
        result = type(self)(self)
        if self._element_type is not None and self._field is not None:
            result.enable_coercing(self._element_type, self._obj, self._field)
        result.symmetric_difference_update(y)
        return result

    def __ixor__(  # type: ignore[override]
        self, y: Iterable[Any]
    ) -> Self:
        return super().__ixor__(self._coerce_iterable(y))


class CompoundFieldType(FieldType[Any], Generic[E]):
    """Base class for compound field types like List, Dict, Set.

    E is the element type stored in the container.
    """

    def __init__(self, element_type: FieldType[E], **field_args: Any) -> None:
        if isinstance(element_type, type):
            raise TypeError(
                _(
                    'Element type must be a FieldType instance, not a class. '
                    'Use %(type)s() instead of %(type)s.'
                )
                % {'type': element_type.__name__}
            )
        # Declare as instance attribute to avoid mypy treating Field as a
        # descriptor (Field implements __get__/__set__ for VersionedObject)
        self._element_type: Field[E] = Field(element_type, **field_args)


class List(CompoundFieldType[E]):
    """A field type for lists of elements of type E."""

    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> list[E]:
        if not isinstance(value, Iterable) or isinstance(
            value, (str, Mapping)
        ):
            raise ValueError(
                _('A list is required in field %(attr)s, not a %(type)s')
                % {'attr': attr, 'type': type(value).__name__}
            )
        coerced_list: CoercedList[E] = CoercedList()
        coerced_list.enable_coercing(self._element_type, obj, attr)
        coerced_list.extend(value)
        return coerced_list

    def to_primitive(
        self, obj: base.VersionedObject, attr: str, value: list[E]
    ) -> list[object]:
        return [self._element_type.to_primitive(obj, attr, x) for x in value]

    def from_primitive(
        self, obj: base.VersionedObject, attr: str, value: Any
    ) -> list[E]:
        return [self._element_type.from_primitive(obj, attr, x) for x in value]

    def stringify(self, value: list[E]) -> str:
        return '[{}]'.format(
            ','.join([self._element_type.stringify(x) for x in value])
        )

    def get_schema(self) -> dict[str, Any]:
        return {
            'type': ['array'],
            'items': self._element_type.get_schema(),
        }


class Dict(CompoundFieldType[E]):
    """A field type for dicts with string keys and values of type E."""

    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> dict[str, E]:
        if not isinstance(value, dict):
            raise ValueError(
                _('A dict is required in field %(attr)s, not a %(type)s')
                % {'attr': attr, 'type': type(value).__name__}
            )
        coerced_dict: CoercedDict[E] = CoercedDict()
        coerced_dict.enable_coercing(self._element_type, obj, attr)
        coerced_dict.update(value)
        return coerced_dict

    def to_primitive(
        self, obj: base.VersionedObject, attr: str, value: dict[str, E]
    ) -> dict[str, object]:
        primitive: dict[str, object] = {}
        for key, element in value.items():
            primitive[key] = self._element_type.to_primitive(
                obj, f'{attr}["{key}"]', element
            )
        return primitive

    def from_primitive(
        self, obj: base.VersionedObject, attr: str, value: Any
    ) -> dict[str, E]:
        concrete: dict[str, E] = {}
        for key, element in value.items():
            concrete[key] = self._element_type.from_primitive(
                obj, f'{attr}["{key}"]', element
            )
        return concrete

    def stringify(self, value: dict[str, E]) -> str:
        return '{{{}}}'.format(
            ','.join(
                [
                    f'{key}={self._element_type.stringify(val)}'
                    for key, val in sorted(value.items())
                ]
            )
        )

    def get_schema(self) -> dict[str, Any]:
        return {
            'type': ['object'],
            'additionalProperties': self._element_type.get_schema(),
        }


class DictProxyField:
    """Descriptor allowing us to assign pinning data as a dict of key_types

    This allows us to have an object field that will be a dict of key_type
    keys, allowing that will convert back to string-keyed dict.

    This will take care of the conversion while the dict field will make sure
    that we store the raw json-serializable data on the object.

    key_type should return a type that unambiguously responds to str
    so that calling key_type on it yields the same thing.
    """

    def __init__(self, dict_field_name: str, key_type: type = int) -> None:
        self._fld_name = dict_field_name
        self._key_type = key_type

    def __get__(
        self, obj: Any, obj_type: type | None = None
    ) -> DictProxyField | dict[Any, Any] | None:
        if obj is None:
            return self
        if getattr(obj, self._fld_name) is None:
            return None
        return {
            self._key_type(k): v
            for k, v in getattr(obj, self._fld_name).items()
        }

    def __set__(self, obj: Any, val: dict[Any, Any] | None) -> None:
        if val is None:
            setattr(obj, self._fld_name, val)
        else:
            setattr(obj, self._fld_name, {str(k): v for k, v in val.items()})


class Set(CompoundFieldType[E]):
    """A field type for sets of elements of type E."""

    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> set[E]:
        if not isinstance(value, set):
            raise ValueError(
                _('A set is required in field %(attr)s, not a %(type)s')
                % {'attr': attr, 'type': type(value).__name__}
            )
        coerced_set: CoercedSet[E] = CoercedSet()
        coerced_set.enable_coercing(self._element_type, obj, attr)
        coerced_set.update(value)
        return coerced_set

    def to_primitive(
        self, obj: base.VersionedObject, attr: str, value: set[E]
    ) -> tuple[object, ...]:
        return tuple(
            self._element_type.to_primitive(obj, attr, x) for x in value
        )

    def from_primitive(
        self, obj: base.VersionedObject, attr: str, value: Any
    ) -> set[E]:
        return {self._element_type.from_primitive(obj, attr, x) for x in value}

    def stringify(self, value: set[E]) -> str:
        return 'set([{}])'.format(
            ','.join([self._element_type.stringify(x) for x in value])
        )

    def get_schema(self) -> dict[str, Any]:
        return {
            'type': ['array'],
            'uniqueItems': True,
            'items': self._element_type.get_schema(),
        }


class Object(FieldType['base.VersionedObject']):
    def __init__(
        self, obj_name: str, subclasses: bool = False, **kwargs: Any
    ) -> None:
        self._obj_name = obj_name
        self._subclasses = subclasses
        super().__init__(**kwargs)

    @staticmethod
    def _get_all_obj_names(obj: Any) -> list[str]:
        obj_names: list[str] = []
        for parent in obj.__class__.mro():
            # Skip mix-ins which are not versioned object subclasses
            if not hasattr(parent, "obj_name"):
                continue
            obj_names.append(parent.obj_name())
        return obj_names

    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> base.VersionedObject:
        try:
            obj_name = value.obj_name()
        except AttributeError:
            obj_name = ""

        if self._subclasses:
            obj_names = self._get_all_obj_names(value)
        else:
            obj_names = [obj_name]

        if self._obj_name not in obj_names:
            if not obj_name:
                # If we're not dealing with an object, it's probably a
                # primitive so get it's type for the message below.
                obj_name = type(value).__name__
            obj_mod = ''
            if hasattr(obj, '__module__'):
                obj_mod = ''.join([obj.__module__, '.'])
            val_mod = ''
            if hasattr(value, '__module__'):
                val_mod = ''.join([value.__module__, '.'])
            raise ValueError(
                _(
                    'An object of type %(type)s is required '
                    'in field %(attr)s, not a %(valtype)s'
                )
                % {
                    'type': ''.join([obj_mod, self._obj_name]),
                    'attr': attr,
                    'valtype': ''.join([val_mod, obj_name]),
                }
            )
        # We've validated value is the correct object type above
        return cast('base.VersionedObject', value)

    def to_primitive(
        self, obj: base.VersionedObject, attr: str, value: base.VersionedObject
    ) -> object:
        return value.obj_to_primitive()

    def from_primitive(
        self, obj: base.VersionedObject, attr: str, value: Any
    ) -> base.VersionedObject:
        # FIXME(danms): Avoid circular import from base.py
        from oslo_versionedobjects import base as obj_base

        # NOTE (ndipanov): If they already got hydrated by the serializer, just
        # pass them back unchanged
        if isinstance(value, obj_base.VersionedObject):
            return value
        return obj.obj_from_primitive(value, obj._context)

    def describe(self) -> str:
        return f"Object<{self._obj_name}>"

    def stringify(self, value: base.VersionedObject) -> str:
        if 'uuid' in value.fields:
            if value.obj_attr_is_set('uuid'):
                assert hasattr(value, 'uuid')  # narrow type
                ident = f'({value.uuid})'
            else:
                ident = '(UNKNOWN)'
        elif 'id' in value.fields:
            if value.obj_attr_is_set('id'):
                assert hasattr(value, 'id')  # narrow type
                ident = f'({value.id})'
            else:
                ident = '(UNKNOWN)'
        else:
            ident = ''

        return f'{value.obj_name()}{ident}'

    def get_schema(self) -> dict[str, Any]:
        from oslo_versionedobjects import base as obj_base

        obj_classes = obj_base.VersionedObjectRegistry.obj_classes()
        if self._obj_name not in obj_classes:
            raise exception.UnsupportedObjectError(objtype=self._obj_name)

        cls = obj_classes[self._obj_name][0]
        namespace_key = cls._obj_primitive_key('namespace')
        name_key = cls._obj_primitive_key('name')
        version_key = cls._obj_primitive_key('version')
        data_key = cls._obj_primitive_key('data')
        changes_key = cls._obj_primitive_key('changes')
        field_schemas = {
            key: field.get_schema() for key, field in cls.fields.items()
        }
        required_fields = [
            key
            for key, field in sorted(cls.fields.items())
            if not field.nullable
        ]
        schema: dict[str, Any] = {
            'type': ['object'],
            'properties': {
                namespace_key: {'type': 'string'},
                name_key: {'type': 'string'},
                version_key: {'type': 'string'},
                changes_key: {
                    'type': 'array',
                    'items': {'type': 'string'},
                },
                data_key: {
                    'type': 'object',
                    'description': f'fields of {self._obj_name}',
                    'properties': field_schemas,
                },
            },
            'required': [namespace_key, name_key, version_key, data_key],
        }

        if required_fields:
            schema['properties'][data_key]['required'] = required_fields

        return schema


class AutoTypedField(Field[T]):
    """Base class for fields with automatic type detection."""

    AUTO_TYPE: FieldType[T]

    def __init__(
        self,
        nullable: bool = False,
        default: T | type[UnspecifiedDefault] = UnspecifiedDefault,
        read_only: bool = False,
    ) -> None:
        super().__init__(
            self.AUTO_TYPE,
            nullable=nullable,
            default=default,
            read_only=read_only,
        )


class StringField(AutoTypedField[str]):
    AUTO_TYPE = String()


class SensitiveStringField(AutoTypedField[str]):
    """Field type that masks passwords when the field is stringified."""

    AUTO_TYPE = SensitiveString()


class VersionPredicateField(AutoTypedField[str]):
    AUTO_TYPE = VersionPredicate()


class BaseEnumField(AutoTypedField[Any]):
    """Base class for all enum field types

    This class should not be directly instantiated. Instead
    subclass it and set AUTO_TYPE to be a SomeEnum()
    where SomeEnum is a subclass of Enum.
    """

    _type: Enum

    def __init__(
        self,
        nullable: bool = False,
        default: str | type[UnspecifiedDefault] = UnspecifiedDefault,
        read_only: bool = False,
    ) -> None:
        if self.AUTO_TYPE is None:
            raise exception.EnumFieldUnset(fieldname=self.__class__.__name__)

        if not isinstance(self.AUTO_TYPE, Enum):
            raise exception.EnumFieldInvalid(
                typename=self.AUTO_TYPE.__class__.__name__,
                fieldname=self.__class__.__name__,
            )

        super().__init__(
            nullable=nullable,
            default=default,
            read_only=read_only,
        )

    def __repr__(self) -> str:
        enum_type = self._type
        valid_values = enum_type.valid_values
        args: dict[str, Any] = {
            'nullable': self._nullable,
            'default': self._default,
            'valid_values': valid_values,
        }
        return '{}({})'.format(
            self._type.__class__.__name__,
            ','.join([f'{k}={v}' for k, v in sorted(args.items())]),
        )

    @property
    def valid_values(self) -> Collection[str]:
        """Return the valid values for the field."""
        enum_type = self._type
        return enum_type.valid_values


class EnumField(BaseEnumField):
    """Anonymous enum field type

    This class allows for anonymous enum types to be
    declared, simply by passing in a list of valid values
    to its constructor. It is generally preferable though,
    to create an explicit named enum type by sub-classing
    the BaseEnumField type directly.
    """

    def __init__(self, valid_values: Collection[Any], **kwargs: Any) -> None:
        self.AUTO_TYPE = Enum(valid_values=valid_values)
        super().__init__(**kwargs)


class StateMachine(EnumField):
    """A mixin that can be applied to an EnumField to enforce a state machine

    e.g: Setting the code below on a field will ensure an object cannot
    transition from ERROR to ACTIVE

    :example:
        .. code-block:: python

            class FakeStateMachineField(fields.EnumField, fields.StateMachine):
                ACTIVE = 'ACTIVE'
                PENDING = 'PENDING'
                ERROR = 'ERROR'
                DELETED = 'DELETED'

                ALLOWED_TRANSITIONS = {
                    ACTIVE: {
                        PENDING,
                        ERROR,
                        DELETED,
                    },
                    PENDING: {ACTIVE, ERROR},
                    ERROR: {
                        PENDING,
                    },
                    DELETED: {},  # This is a terminal state
                }

                _TYPES = (ACTIVE, PENDING, ERROR, DELETED)

                def __init__(self, **kwargs):
                    super(FakeStateMachineField, self).__init__(
                        self._TYPES, **kwargs
                    )

    """

    # This is dict of states, that have dicts of states an object is
    # allowed to transition to

    ALLOWED_TRANSITIONS: dict[str, set[str]] = {}

    def _my_name(self, obj: base.VersionedObject) -> str:
        for name, field in obj.fields.items():
            if field == self:
                return name
        return 'unknown'

    def coerce(
        self, obj: base.VersionedObject | None, attr: str, value: Any
    ) -> str:
        super().coerce(obj, attr, value)

        if not obj or attr not in obj:
            return str(value)

        my_name = self._my_name(obj)
        msg = _(
            "%(object)s.%(name)s is not allowed to transition out of "
            "%(value)s state"
        )

        current_value = getattr(obj, attr)
        if current_value in self.ALLOWED_TRANSITIONS:
            if value in self.ALLOWED_TRANSITIONS[current_value]:
                return str(value)

            msg = _(
                "%(object)s.%(name)s is not allowed to transition out of "
                "'%(current_value)s' state to '%(value)s' state, choose "
                "from %(options)r"
            )
        msg = msg % {
            'object': obj.obj_name(),
            'name': my_name,
            'current_value': current_value,
            'value': value,
            'options': [x for x in self.ALLOWED_TRANSITIONS[current_value]],
        }
        raise ValueError(msg)


class UUIDField(AutoTypedField[str]):
    """UUID Field Type

    .. warning::

        This class does not actually validate UUIDs. This will happen in a
        future major version of oslo.versionedobjects

    To validate that you have valid UUIDs you need to do the following in
    your own objects/fields.py

    :Example:
        .. code-block:: python

            import oslo_versionedobjects.fields as ovo_fields


            class UUID(ovo_fields.UUID):
                def coerce(self, obj, attr, value):
                    uuid.UUID(value)
                    return str(value)


            class UUIDField(ovo_fields.AutoTypedField):
                AUTO_TYPE = UUID()

    and then in your objects use
    ``<your_projects>.object.fields.UUIDField``.

    This will become default behaviour in the future.
    """

    AUTO_TYPE = UUID()


class MACAddressField(AutoTypedField[str]):
    AUTO_TYPE = MACAddress()


class PCIAddressField(AutoTypedField[str]):
    AUTO_TYPE = PCIAddress()


class IntegerField(AutoTypedField[int]):
    AUTO_TYPE = Integer()


class NonNegativeIntegerField(AutoTypedField[int]):
    AUTO_TYPE = NonNegativeInteger()


class FloatField(AutoTypedField[float]):
    AUTO_TYPE = Float()


class NonNegativeFloatField(AutoTypedField[float]):
    AUTO_TYPE = NonNegativeFloat()


# This is a strict interpretation of boolean
# values using Python's semantics for truth/falsehood
class BooleanField(AutoTypedField[bool]):
    AUTO_TYPE = Boolean()


# This is a flexible interpretation of boolean
# values using common user friendly semantics for
# truth/falsehood. ie strings like 'yes', 'no',
# 'on', 'off', 't', 'f' get mapped to values you
# would expect.
class FlexibleBooleanField(AutoTypedField[bool]):
    AUTO_TYPE = FlexibleBoolean()


class DateTimeField(AutoTypedField[datetime.datetime]):
    def __init__(self, tzinfo_aware: bool = True, **kwargs: Any) -> None:
        self.AUTO_TYPE = DateTime(tzinfo_aware=tzinfo_aware)
        super().__init__(**kwargs)


class DictOfStringsField(AutoTypedField[dict[str, str]]):
    AUTO_TYPE = Dict(String())


class DictOfNullableStringsField(AutoTypedField[dict[str, str | None]]):
    AUTO_TYPE = Dict(String(), nullable=True)


class DictOfIntegersField(AutoTypedField[dict[str, int]]):
    AUTO_TYPE = Dict(Integer())


class ListOfStringsField(AutoTypedField[list[str]]):
    AUTO_TYPE = List(String())


class SetOfStringsField(AutoTypedField[set[str]]):
    AUTO_TYPE = Set(String())


class ListOfListsOfStringsField(AutoTypedField[list[list[str]]]):
    AUTO_TYPE = List(List(String()))


class DictOfListOfStringsField(AutoTypedField[dict[str, list[str]]]):
    AUTO_TYPE = Dict(List(String()))


class DictOfSetOfIntegersField(AutoTypedField[dict[str, set[int]]]):
    AUTO_TYPE = Dict(Set(Integer()))


class ListOfEnumField(AutoTypedField[list[str]]):
    def __init__(self, valid_values: list[str], **kwargs: Any) -> None:
        self._valid_values = valid_values
        self.AUTO_TYPE = List(Enum(valid_values))
        super().__init__(**kwargs)

    def __repr__(self) -> str:
        args: dict[str, Any] = {
            'nullable': self._nullable,
            'default': self._default,
            'valid_values': self._valid_values,
        }
        return '{}({})'.format(
            self._type.__class__.__name__,
            ','.join([f'{k}={v}' for k, v in sorted(args.items())]),
        )


class SetOfIntegersField(AutoTypedField[set[int]]):
    AUTO_TYPE = Set(Integer())


class ListOfSetsOfIntegersField(AutoTypedField[list[set[int]]]):
    AUTO_TYPE = List(Set(Integer()))


class ListOfIntegersField(AutoTypedField[list[int]]):
    AUTO_TYPE = List(Integer())


class ListOfDictOfNullableStringsField(
    AutoTypedField[list[dict[str, str | None]]]
):
    AUTO_TYPE = List(Dict(String(), nullable=True))


class ObjectField(AutoTypedField['base.VersionedObject']):
    def __init__(
        self, objtype: str, subclasses: bool = False, **kwargs: Any
    ) -> None:
        self.AUTO_TYPE = Object(objtype, subclasses)
        self.objname = objtype
        super().__init__(**kwargs)


class ListOfObjectsField(AutoTypedField[list['base.VersionedObject']]):
    def __init__(
        self, objtype: str, subclasses: bool = False, **kwargs: Any
    ) -> None:
        self.AUTO_TYPE = List(Object(objtype, subclasses))
        self.objname = objtype
        super().__init__(**kwargs)


class ListOfUUIDField(AutoTypedField[list[str]]):
    AUTO_TYPE = List(UUID())


class IPAddressField(AutoTypedField[netaddr.IPAddress]):
    AUTO_TYPE = IPAddress()


class IPV4AddressField(AutoTypedField[netaddr.IPAddress]):
    AUTO_TYPE = IPV4Address()


class IPV6AddressField(AutoTypedField[netaddr.IPAddress]):
    AUTO_TYPE = IPV6Address()


class IPV4AndV6AddressField(AutoTypedField[netaddr.IPAddress]):
    AUTO_TYPE = IPV4AndV6Address()


class IPNetworkField(AutoTypedField[netaddr.IPNetwork]):
    AUTO_TYPE = IPNetwork()


class IPV4NetworkField(AutoTypedField[netaddr.IPNetwork]):
    AUTO_TYPE = IPV4Network()


class IPV6NetworkField(AutoTypedField[netaddr.IPNetwork]):
    AUTO_TYPE = IPV6Network()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from oslo_config import cfg

from oslo_versionedobjects import exception


def list_opts() -> list[tuple[str, list[cfg.Opt]]]:
    """Returns a list of oslo.config options available in the library.

    The returned list includes all oslo.config options which may be registered
    at runtime by the library.

    Each element of the list is a tuple. The first element is the name of the
    group under which the list of elements in the second element will be
    registered. A group name of None corresponds to the [DEFAULT] group in
    config files.

    The purpose of this is to allow tools like the Oslo sample config file
    generator to discover the options exposed to users by this library.

    :returns: a list of (group_name, opts) tuples
    """
    return [('oslo_versionedobjects', copy.deepcopy(exception.exc_log_opts))]



#    Copyright 2013 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from typing import Any
from unittest import mock
import warnings

import netaddr
import testtools

from oslo_versionedobjects import _utils
from oslo_versionedobjects import base as obj_base
from oslo_versionedobjects import exception
from oslo_versionedobjects import fields
from oslo_versionedobjects import test


class FakeFieldType(fields.FieldType[str]):
    def coerce(self, obj, attr, value):
        return f'*{value}*'

    def to_primitive(self, obj, attr, value):
        return f'!{value}!'

    def from_primitive(self, obj, attr, value):
        return value[1:-1]

    def get_schema(self):
        return {'type': ['foo']}


class FakeEnum(fields.Enum):
    FROG = "frog"
    PLATYPUS = "platypus"
    ALLIGATOR = "alligator"

    ALL = (FROG, PLATYPUS, ALLIGATOR)

    def __init__(self, **kwargs):
        super().__init__(valid_values=FakeEnum.ALL, **kwargs)


class FakeEnumAlt(fields.Enum):
    FROG = "frog"
    PLATYPUS = "platypus"
    AARDVARK = "aardvark"

    ALL = {FROG, PLATYPUS, AARDVARK}

    def __init__(self, **kwargs):
        super().__init__(valid_values=FakeEnumAlt.ALL, **kwargs)


class FakeEnumField(fields.BaseEnumField):
    AUTO_TYPE = FakeEnum()


class FakeStateMachineField(fields.StateMachine):
    ACTIVE = 'ACTIVE'
    PENDING = 'PENDING'
    ERROR = 'ERROR'

    ALLOWED_TRANSITIONS = {
        ACTIVE: {PENDING, ERROR},
        PENDING: {ACTIVE, ERROR},
        ERROR: {
            PENDING,
        },
    }

    _TYPES = (ACTIVE, PENDING, ERROR)

    def __init__(self, **kwargs):
        super().__init__(self._TYPES, **kwargs)


class FakeEnumAltField(fields.BaseEnumField):
    AUTO_TYPE = FakeEnumAlt()


class TestFieldType(test.TestCase):
    def test_get_schema(self):
        self.assertRaises(NotImplementedError, fields.FieldType().get_schema)


class TestField(test.TestCase):
    def setUp(self):
        super().setUp()
        self.field: Any = fields.Field(FakeFieldType())
        self.coerce_good_values: list[tuple[Any, Any]] = [('foo', '*foo*')]
        self.coerce_bad_values: list[Any] = []
        self.to_primitive_values: list[tuple[Any, Any]] = [('foo', '!foo!')]
        self.from_primitive_values: list[tuple[Any, Any]] = [('!foo!', 'foo')]

    def test_coerce_good_values(self):
        for in_val, out_val in self.coerce_good_values:
            self.assertEqual(out_val, self.field.coerce(None, 'attr', in_val))

    def test_coerce_bad_values(self):
        for in_val in self.coerce_bad_values:
            self.assertRaises(
                (TypeError, ValueError),
                self.field.coerce,
                None,
                'attr',
                in_val,
            )

    def test_to_primitive(self):
        for in_val, prim_val in self.to_primitive_values:
            self.assertEqual(
                prim_val, self.field.to_primitive('obj', 'attr', in_val)
            )

    def test_from_primitive(self):
        class AnObject(obj_base.VersionedObject):
            fields = {
                'intfield': fields.IntegerField(),
            }

        for prim_val, out_val in self.from_primitive_values:
            obj = AnObject(intfield=5)
            self.assertEqual(
                out_val, self.field.from_primitive(obj, 'attr', prim_val)
            )

    def test_stringify(self):
        self.assertEqual('123', self.field.stringify(123))


class TestSchema(test.TestCase):
    def setUp(self):
        super().setUp()
        self.field = fields.Field(
            FakeFieldType(), nullable=True, default='', read_only=False
        )

    def test_get_schema(self):
        self.assertEqual(
            {'type': ['foo', 'null'], 'default': '', 'readonly': False},
            self.field.get_schema(),
        )


class TestString(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.StringField()
        self.coerce_good_values = [
            ('foo', 'foo'),
            (1, '1'),
            (1.0, '1.0'),
            (True, 'True'),
        ]
        self.coerce_bad_values = [None]
        self.to_primitive_values = self.coerce_good_values[0:1]
        self.from_primitive_values = self.coerce_good_values[0:1]

    def test_stringify(self):
        self.assertEqual("'123'", self.field.stringify(123))

    def test_fieldtype_get_schema(self):
        self.assertEqual({'type': ['string']}, self.field._type.get_schema())

    def test_get_schema(self):
        self.assertEqual(
            {'type': ['string'], 'readonly': False}, self.field.get_schema()
        )


class TestSensitiveString(TestString):
    def setUp(self):
        super().setUp()
        self.field = fields.SensitiveStringField()

    def test_stringify(self):
        payload = """{'admin_password':'mypassword'}"""
        expected = """'{'admin_password':'***'}'"""
        self.assertEqual(expected, self.field.stringify(payload))


class TestVersionPredicate(TestString):
    def setUp(self):
        super().setUp()
        self.field = fields.VersionPredicateField()
        self.coerce_good_values = [
            ('>=1.0', '>=1.0'),
            ('>1.0', '>1.0'),
            ('==1.1', '==1.1'),
            ('<=1.1.0', '<=1.1.0'),
            ('<1.1.0', '<1.1.0'),
        ]
        self.coerce_bad_values = ['1', 'foo', 1.0, '1.0', '=1.0']
        self.to_primitive_values = self.coerce_good_values[0:1]
        self.from_primitive_values = self.coerce_good_values[0:1]


class TestMACAddress(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.MACAddressField()
        self.coerce_good_values = [
            ('c6:df:11:a5:c8:5d', 'c6:df:11:a5:c8:5d'),
            ('C6:DF:11:A5:C8:5D', 'c6:df:11:a5:c8:5d'),
            ('c6:df:11:a5:c8:5d', 'c6:df:11:a5:c8:5d'),
            ('C6:DF:11:A5:C8:5D', 'c6:df:11:a5:c8:5d'),
        ]
        self.coerce_bad_values = [
            'C6:DF:11:A5:C8',  # Too short
            'C6:DF:11:A5:C8:5D:D7',  # Too long
            'C6:DF:11:A5:C8:KD',  # Bad octal
            1123123,  # Number
            {},  # dict
        ]
        self.to_primitive_values = self.coerce_good_values[0:1]
        self.from_primitive_values = self.coerce_good_values[0:1]

    def test_get_schema(self):
        schema = self.field.get_schema()
        self.assertEqual(['string'], schema['type'])
        self.assertEqual(False, schema['readonly'])
        pattern = schema['pattern']
        for _, valid_val in self.coerce_good_values:
            self.assertRegex(valid_val, pattern)
        invalid_vals = [
            x for x in self.coerce_bad_values if isinstance(x, str)
        ]
        for invalid_val in invalid_vals:
            self.assertNotRegex(invalid_val, pattern)


class TestPCIAddress(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.PCIAddressField()
        self.coerce_good_values = [
            ('0000:02:00.0', '0000:02:00.0'),
            ('FFFF:FF:1F.7', 'ffff:ff:1f.7'),
            ('fFfF:fF:1F.7', 'ffff:ff:1f.7'),
        ]
        self.coerce_bad_values = [
            '000:02:00.0',  # Too short
            '00000:02:00.0',  # Too long
            'FFFF:FF:2F.7',  # Bad slot
            'FFFF:GF:1F.7',  # Bad octal
            1123123,  # Number
            {},  # dict
        ]
        self.to_primitive_values = self.coerce_good_values[0:1]
        self.from_primitive_values = self.coerce_good_values[0:1]

    def test_get_schema(self):
        schema = self.field.get_schema()
        self.assertEqual(['string'], schema['type'])
        self.assertEqual(False, schema['readonly'])
        pattern = schema['pattern']
        for _, valid_val in self.coerce_good_values:
            self.assertRegex(valid_val, pattern)
        invalid_vals = [
            x for x in self.coerce_bad_values if isinstance(x, str)
        ]
        for invalid_val in invalid_vals:
            self.assertNotRegex(invalid_val, pattern)


class TestUUID(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.UUIDField()
        self.coerce_good_values = [
            (
                'da66a411-af0e-4829-9b67-475017ddd152',
                'da66a411-af0e-4829-9b67-475017ddd152',
            ),
            (
                'da66a411af0e48299b67475017ddd152',
                'da66a411af0e48299b67475017ddd152',
            ),
            (
                'DA66A411-AF0E-4829-9B67-475017DDD152',
                'DA66A411-AF0E-4829-9B67-475017DDD152',
            ),
            (
                'DA66A411AF0E48299b67475017DDD152',
                'DA66A411AF0E48299b67475017DDD152',
            ),
            # These values are included to ensure there is not change in
            # behaviour - only when we can remove the old UUID behaviour can
            #  we add these to the "self.coerce_bad_values" list
            ('da66a411-af0e-4829-9b67', 'da66a411-af0e-4829-9b67'),
            (
                'da66a411-af0e-4829-9b67-475017ddd152548999',
                'da66a411-af0e-4829-9b67-475017ddd152548999',
            ),
            (
                'da66a411-af0e-4829-9b67-475017ddz152',
                'da66a411-af0e-4829-9b67-475017ddz152',
            ),
            ('fake_uuid', 'fake_uuid'),
            ('fake_uāid', 'fake_uāid'),
            (
                b'fake_u\xe1id'.decode('latin_1'),
                b'fake_u\xe1id'.decode('latin_1'),
            ),
            ('1', '1'),
            (1, '1'),
        ]
        self.to_primitive_values = self.coerce_good_values[0:1]
        self.from_primitive_values = self.coerce_good_values[0:1]

    @mock.patch('warnings.warn')
    def test_coerce_good_values(self, mock_warn):
        super().test_coerce_good_values()
        mock_warn.assert_has_calls(
            [mock.call(mock.ANY, FutureWarning)] * 8,
        )

    def test_validation_warning_can_be_escalated_to_exception(self):
        warnings.filterwarnings(action='error')
        self.assertRaises(
            FutureWarning, self.field.coerce, 'obj', 'attr', 'not a uuid'
        )

    def test_get_schema(self):
        field = fields.UUIDField()
        schema = field.get_schema()
        self.assertEqual(['string'], schema['type'])
        self.assertEqual(False, schema['readonly'])
        pattern = schema['pattern']
        for _, valid_val in self.coerce_good_values[:4]:
            self.assertRegex(valid_val, pattern)
        invalid_vals = [
            x for x in self.coerce_bad_values if isinstance(x, str)
        ]
        for invalid_val in invalid_vals:
            self.assertNotRegex(invalid_val, pattern)


class TestBaseEnum(TestField):
    def setUp(self):
        super().setUp()
        self.field = FakeEnumField()
        self.coerce_good_values = [
            ('frog', 'frog'),
            ('platypus', 'platypus'),
            ('alligator', 'alligator'),
        ]
        self.coerce_bad_values = ['aardvark', 'wookie']
        self.to_primitive_values = self.coerce_good_values[0:1]
        self.from_primitive_values = self.coerce_good_values[0:1]

    def test_stringify(self):
        self.assertEqual("'platypus'", self.field.stringify('platypus'))

    def test_stringify_invalid(self):
        self.assertRaises(ValueError, self.field.stringify, 'aardvark')

    def test_fingerprint(self):
        # Notes(yjiang5): make sure changing valid_value will be detected
        # in test_objects.test_versions
        field1 = FakeEnumField()
        field2 = FakeEnumAltField()
        self.assertNotEqual(str(field1), str(field2))

    def test_valid_values(self):
        self.assertEqual(self.field.valid_values, FakeEnum.ALL)

    def test_valid_values_keeps_type(self):
        self.assertIsInstance(self.field.valid_values, tuple)
        self.assertIsInstance(FakeEnumAltField().valid_values, set)


class TestEnum(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.EnumField(valid_values=['foo', 'bar', 1, True])
        self.coerce_good_values = [('foo', 'foo'), (1, '1'), (True, 'True')]
        self.coerce_bad_values = ['boo', 2, False]
        self.to_primitive_values = self.coerce_good_values[0:1]
        self.from_primitive_values = self.coerce_good_values[0:1]

    def test_stringify(self):
        self.assertEqual("'foo'", self.field.stringify('foo'))

    def test_stringify_invalid(self):
        self.assertRaises(ValueError, self.field.stringify, '123')

    def test_fieldtype_get_schema(self):
        self.assertEqual(
            {'type': ['string'], 'enum': ["foo", "bar", 1, True]},
            self.field._type.get_schema(),
        )

    def test_get_schema(self):
        self.assertEqual(
            {
                'type': ['string'],
                'enum': ["foo", "bar", 1, True],
                'readonly': False,
            },
            self.field.get_schema(),
        )

    def test_fingerprint(self):
        # Notes(yjiang5): make sure changing valid_value will be detected
        # in test_objects.test_versions
        field1 = fields.EnumField(valid_values=['foo', 'bar'])
        field2 = fields.EnumField(valid_values=['foo', 'bar1'])
        self.assertNotEqual(str(field1), str(field2))

    def test_missing_valid_values(self):
        self.assertRaises(
            exception.EnumRequiresValidValuesError, fields.EnumField, None
        )

    def test_empty_valid_values(self):
        self.assertRaises(
            exception.EnumRequiresValidValuesError, fields.EnumField, []
        )

    def test_non_iterable_valid_values(self):
        self.assertRaises(
            exception.EnumValidValuesInvalidError, fields.EnumField, True
        )

    def test_enum_subclass_check(self):
        def _test():
            class BrokenEnumField(fields.BaseEnumField):
                # this is intentionally using the wrong type
                AUTO_TYPE = int  # type: ignore[assignment]

            BrokenEnumField()

        self.assertRaises(exception.EnumFieldInvalid, _test)


class TestStateMachine(TestField):
    def test_good_transitions(self):
        @obj_base.VersionedObjectRegistry.register
        class AnObject(obj_base.VersionedObject):
            fields = {
                'status': FakeStateMachineField(),
            }

        obj = AnObject()

        obj.status = FakeStateMachineField.ACTIVE
        obj.status = FakeStateMachineField.PENDING
        obj.status = FakeStateMachineField.ERROR
        obj.status = FakeStateMachineField.PENDING
        obj.status = FakeStateMachineField.ACTIVE

    def test_bad_transitions(self):
        @obj_base.VersionedObjectRegistry.register
        class AnObject(obj_base.VersionedObject):
            fields = {
                'status': FakeStateMachineField(),
            }

        obj = AnObject(status='ERROR')

        try:
            obj.status = FakeStateMachineField.ACTIVE
        except ValueError as ex:
            self.assertEqual(
                'AnObject.status is not allowed to transition out '
                'of \'ERROR\' state to \'ACTIVE\' state, choose from '
                '[\'PENDING\']',
                str(ex),
            )
        else:
            raise Exception('Invalid transition failed to raise error')

    def test_bad_initial_value(self):
        @obj_base.VersionedObjectRegistry.register
        class AnObject(obj_base.VersionedObject):
            fields = {
                'status': FakeStateMachineField(),
            }

        obj = AnObject()

        with testtools.ExpectedException(ValueError):
            obj.status = "FOO"

    def test_bad_updated_value(self):
        @obj_base.VersionedObjectRegistry.register
        class AnObject(obj_base.VersionedObject):
            fields = {
                'status': FakeStateMachineField(),
            }

        obj = AnObject()

        with testtools.ExpectedException(ValueError):
            obj.status = FakeStateMachineField.ACTIVE
            obj.status = "FOO"


class TestInteger(TestField):
    def setUp(self):
        super(TestField, self).setUp()
        self.field = fields.IntegerField()
        self.coerce_good_values = [(1, 1), ('1', 1)]
        self.coerce_bad_values = ['foo', None]
        self.to_primitive_values = self.coerce_good_values[0:1]
        self.from_primitive_values = self.coerce_good_values[0:1]

    def test_fieldtype_get_schema(self):
        self.assertEqual({'type': ['integer']}, self.field._type.get_schema())

    def test_get_schema(self):
        self.assertEqual(
            {'type': ['integer'], 'readonly': False}, self.field.get_schema()
        )


class TestNonNegativeInteger(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.NonNegativeIntegerField()
        self.coerce_good_values = [(1, 1), ('1', 1)]
        self.coerce_bad_values = ['-2', '4.2', 'foo', None]
        self.to_primitive_values = self.coerce_good_values[0:1]
        self.from_primitive_values = self.coerce_good_values[0:1]

    def test_get_schema(self):
        self.assertEqual(
            {'type': ['integer'], 'readonly': False, 'minimum': 0},
            self.field.get_schema(),
        )


class TestFloat(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.FloatField()
        self.coerce_good_values = [(1.1, 1.1), ('1.1', 1.1)]
        self.coerce_bad_values = ['foo', None]
        self.to_primitive_values = self.coerce_good_values[0:1]
        self.from_primitive_values = self.coerce_good_values[0:1]

    def test_fieldtype_get_schema(self):
        self.assertEqual({'type': ['number']}, self.field._type.get_schema())

    def test_get_schema(self):
        self.assertEqual(
            {'type': ['number'], 'readonly': False}, self.field.get_schema()
        )


class TestNonNegativeFloat(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.NonNegativeFloatField()
        self.coerce_good_values = [(1.1, 1.1), ('1.1', 1.1)]
        self.coerce_bad_values = ['-4.2', 'foo', None]
        self.to_primitive_values = self.coerce_good_values[0:1]
        self.from_primitive_values = self.coerce_good_values[0:1]

    def test_get_schema(self):
        self.assertEqual(
            {'type': ['number'], 'readonly': False, 'minimum': 0},
            self.field.get_schema(),
        )


class TestBoolean(TestField):
    def setUp(self):
        super(TestField, self).setUp()
        self.field = fields.BooleanField()
        self.coerce_good_values = [
            (True, True),
            (False, False),
            (1, True),
            ('foo', True),
            (0, False),
            ('', False),
        ]
        self.coerce_bad_values = []
        self.to_primitive_values = self.coerce_good_values[0:2]
        self.from_primitive_values = self.coerce_good_values[0:2]

    def test_fieldtype_get_schema(self):
        self.assertEqual({'type': ['boolean']}, self.field._type.get_schema())

    def test_get_schema(self):
        self.assertEqual(
            {'type': ['boolean'], 'readonly': False}, self.field.get_schema()
        )


class TestFlexibleBoolean(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.FlexibleBooleanField()
        self.coerce_good_values = [
            (True, True),
            (False, False),
            ("true", True),
            ("false", False),
            ("t", True),
            ("f", False),
            ("yes", True),
            ("no", False),
            ("y", True),
            ("n", False),
            ("on", True),
            ("off", False),
            (1, True),
            (0, False),
            ('frog', False),
            ('', False),
        ]
        self.coerce_bad_values = []
        self.to_primitive_values = self.coerce_good_values[0:2]
        self.from_primitive_values = self.coerce_good_values[0:2]


class TestDateTime(TestField):
    def setUp(self):
        super().setUp()
        self.dt = datetime.datetime(1955, 11, 5, tzinfo=datetime.UTC)
        self.field = fields.DateTimeField()
        self.coerce_good_values = [
            (self.dt, self.dt),
            (_utils.isotime(self.dt), self.dt),
        ]
        self.coerce_bad_values = [1, 'foo']
        self.to_primitive_values = [(self.dt, _utils.isotime(self.dt))]
        self.from_primitive_values = [(_utils.isotime(self.dt), self.dt)]

    def test_stringify(self):
        self.assertEqual(
            '1955-11-05T18:00:00Z',
            self.field.stringify(
                datetime.datetime(1955, 11, 5, 18, 0, 0, tzinfo=datetime.UTC)
            ),
        )

    def test_get_schema(self):
        self.assertEqual(
            {'type': ['string'], 'format': 'date-time', 'readonly': False},
            self.field.get_schema(),
        )


class TestDateTimeNoTzinfo(TestField):
    def setUp(self):
        super().setUp()
        self.dt = datetime.datetime(1955, 11, 5)
        self.field = fields.DateTimeField(tzinfo_aware=False)
        self.coerce_good_values = [
            (self.dt, self.dt),
            (_utils.isotime(self.dt), self.dt),
        ]
        self.coerce_bad_values = [1, 'foo']
        self.to_primitive_values = [(self.dt, _utils.isotime(self.dt))]
        self.from_primitive_values = [
            (
                _utils.isotime(self.dt),
                self.dt,
            )
        ]

    def test_stringify(self):
        self.assertEqual(
            '1955-11-05T18:00:00Z',
            self.field.stringify(datetime.datetime(1955, 11, 5, 18, 0, 0)),
        )


class TestDict(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.Field(fields.Dict(FakeFieldType()))
        self.coerce_good_values = [
            ({'foo': 'bar'}, {'foo': '*bar*'}),
            ({'foo': 1}, {'foo': '*1*'}),
        ]
        self.coerce_bad_values = [{1: 'bar'}, 'foo']
        self.to_primitive_values = [({'foo': 'bar'}, {'foo': '!bar!'})]
        self.from_primitive_values = [({'foo': '!bar!'}, {'foo': 'bar'})]

    def test_stringify(self):
        self.assertEqual("{key=val}", self.field.stringify({'key': 'val'}))

    def test_get_schema(self):
        self.assertEqual(
            {
                'type': ['object'],
                'additionalProperties': {'readonly': False, 'type': ['foo']},
                'readonly': False,
            },
            self.field.get_schema(),
        )


class TestDictOfStrings(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.DictOfStringsField()
        self.coerce_good_values = [
            ({'foo': 'bar'}, {'foo': 'bar'}),
            ({'foo': 1}, {'foo': '1'}),
        ]
        self.coerce_bad_values = [{1: 'bar'}, {'foo': None}, 'foo']
        self.to_primitive_values = [({'foo': 'bar'}, {'foo': 'bar'})]
        self.from_primitive_values = [({'foo': 'bar'}, {'foo': 'bar'})]

    def test_stringify(self):
        self.assertEqual("{key='val'}", self.field.stringify({'key': 'val'}))


class TestDictOfIntegers(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.DictOfIntegersField()
        self.coerce_good_values = [
            ({'foo': '42'}, {'foo': 42}),
            ({'foo': 4.2}, {'foo': 4}),
        ]
        self.coerce_bad_values = [
            {1: 'bar'},
            {'foo': 'boo'},
            'foo',
            {'foo': None},
        ]
        self.to_primitive_values = [({'foo': 42}, {'foo': 42})]
        self.from_primitive_values = [({'foo': 42}, {'foo': 42})]

    def test_stringify(self):
        self.assertEqual("{key=42}", self.field.stringify({'key': 42}))


class TestDictOfStringsNone(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.DictOfNullableStringsField()
        self.coerce_good_values = [
            ({'foo': 'bar'}, {'foo': 'bar'}),
            ({'foo': 1}, {'foo': '1'}),
            ({'foo': None}, {'foo': None}),
        ]
        self.coerce_bad_values = [{1: 'bar'}, 'foo']
        self.to_primitive_values = [({'foo': 'bar'}, {'foo': 'bar'})]
        self.from_primitive_values = [({'foo': 'bar'}, {'foo': 'bar'})]

    def test_stringify(self):
        self.assertEqual(
            "{k2=None,key='val'}",
            self.field.stringify({'k2': None, 'key': 'val'}),
        )


class TestListOfDictOfNullableStringsField(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.ListOfDictOfNullableStringsField()
        self.coerce_good_values = [
            (
                [{'f': 'b', 'f1': 'b1'}, {'f2': 'b2'}],
                [{'f': 'b', 'f1': 'b1'}, {'f2': 'b2'}],
            ),
            ([{'f': 1}, {'f1': 'b1'}], [{'f': '1'}, {'f1': 'b1'}]),
            ([{'foo': None}], [{'foo': None}]),
        ]
        self.coerce_bad_values = [[{1: 'a'}], ['ham', 1], ['eggs']]
        self.to_primitive_values = [
            (
                [{'f': 'b'}, {'f1': 'b1'}, {'f2': None}],
                [{'f': 'b'}, {'f1': 'b1'}, {'f2': None}],
            )
        ]
        self.from_primitive_values = [
            (
                [{'f': 'b'}, {'f1': 'b1'}, {'f2': None}],
                [{'f': 'b'}, {'f1': 'b1'}, {'f2': None}],
            )
        ]

    def test_stringify(self):
        self.assertEqual(
            "[{f=None,f1='b1'},{f2='b2'}]",
            self.field.stringify([{'f': None, 'f1': 'b1'}, {'f2': 'b2'}]),
        )


class TestList(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.Field(fields.List(FakeFieldType()))
        self.coerce_good_values = [(['foo', 'bar'], ['*foo*', '*bar*'])]
        self.coerce_bad_values = ['foo']
        self.to_primitive_values = [(['foo'], ['!foo!'])]
        self.from_primitive_values = [(['!foo!'], ['foo'])]

    def test_stringify(self):
        self.assertEqual('[123]', self.field.stringify([123]))

    def test_fieldtype_get_schema(self):
        self.assertEqual(
            {'type': ['array'], 'items': {'type': ['foo'], 'readonly': False}},
            self.field._type.get_schema(),
        )

    def test_get_schema(self):
        self.assertEqual(
            {
                'type': ['array'],
                'items': {'type': ['foo'], 'readonly': False},
                'readonly': False,
            },
            self.field.get_schema(),
        )


class TestListOfStrings(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.ListOfStringsField()
        self.coerce_good_values = [(['foo', 'bar'], ['foo', 'bar'])]
        self.coerce_bad_values = ['foo']
        self.to_primitive_values = [(['foo'], ['foo'])]
        self.from_primitive_values = [(['foo'], ['foo'])]

    def test_stringify(self):
        self.assertEqual("['abc']", self.field.stringify(['abc']))


class TestSetOfStrings(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.SetOfStringsField()
        self.coerce_good_values = [
            ({'foo', 'bar'}, {'foo', 'bar'}),
            ({'foo', 1}, {'foo', '1'}),
        ]
        self.coerce_bad_values = [['foo']]
        self.to_primitive_values = [({'foo'}, tuple(['foo']))]
        self.from_primitive_values = [(tuple(['foo']), {'foo'})]

    def test_stringify(self):
        self.assertEqual("set(['abc'])", self.field.stringify({'abc'}))


class TestListOfListsOfStrings(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.ListOfListsOfStringsField()
        self.coerce_good_values = [
            ([['foo', 'bar']], [['foo', 'bar']]),
            ([['foo', 1]], [['foo', '1']]),
        ]
        self.coerce_bad_values = ['foo', [['foo', None]]]
        self.to_primitive_values = [([['foo']], [['foo']])]
        self.from_primitive_values = [([['foo']], [['foo']])]

    def test_stringify(self):
        self.assertEqual("[['abc']]", self.field.stringify([['abc']]))


class TestDictOfListOfStrings(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.DictOfListOfStringsField()
        self.coerce_good_values = [
            ({'foo': ['1', '2']}, {'foo': ['1', '2']}),
            ({'foo': [1]}, {'foo': ['1']}),
        ]
        self.coerce_bad_values = [{'foo': [None, None]}, 'foo']
        self.to_primitive_values = [({'foo': ['1', '2']}, {'foo': ['1', '2']})]
        self.from_primitive_values = [
            ({'foo': ['1', '2']}, {'foo': ['1', '2']})
        ]

    def test_stringify(self):
        self.assertEqual(
            "{foo=['1','2']}", self.field.stringify({'foo': ['1', '2']})
        )


class TestDictOfSetOfIntegers(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.DictOfSetOfIntegersField()
        self.coerce_good_values = [({'foo': {'1', 2}}, {'foo': {1, 2}})]
        self.coerce_bad_values = [{'foo': {'bar'}}]
        self.to_primitive_values = [({'foo': {1}}, {'foo': tuple([1])})]
        self.from_primitive_values = [({'foo': tuple([1])}, {'foo': {1}})]

    def test_stringify(self):
        self.assertEqual('{foo=set([1])}', self.field.stringify({'foo': {1}}))


class TestListOfEnum(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.ListOfEnumField(valid_values=['foo', 'bar'])
        self.coerce_good_values = [(['foo', 'bar'], ['foo', 'bar'])]
        self.coerce_bad_values = ['foo', ['foo', 'bar1']]
        self.to_primitive_values = [(['foo'], ['foo'])]
        self.from_primitive_values = [(['foo'], ['foo'])]

    def test_stringify(self):
        self.assertEqual("['foo']", self.field.stringify(['foo']))

    def test_stringify_invalid(self):
        self.assertRaises(ValueError, self.field.stringify, '[abc]')

    def test_fingerprint(self):
        # Notes(yjiang5): make sure changing valid_value will be detected
        # in test_objects.test_versions
        field1 = fields.ListOfEnumField(valid_values=['foo', 'bar'])
        field2 = fields.ListOfEnumField(valid_values=['foo', 'bar1'])
        self.assertNotEqual(str(field1), str(field2))


class TestSet(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.Field(fields.Set(FakeFieldType()))
        self.coerce_good_values = [({'foo', 'bar'}, {'*foo*', '*bar*'})]
        self.coerce_bad_values = [['foo'], {'foo': 'bar'}]
        self.to_primitive_values = [({'foo'}, tuple(['!foo!']))]
        self.from_primitive_values = [(tuple(['!foo!']), {'foo'})]

    def test_stringify(self):
        self.assertEqual('set([123])', self.field.stringify({123}))

    def test_get_schema(self):
        self.assertEqual(
            {
                'type': ['array'],
                'uniqueItems': True,
                'items': {'type': ['foo'], 'readonly': False},
                'readonly': False,
            },
            self.field.get_schema(),
        )


class TestSetOfIntegers(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.SetOfIntegersField()
        self.coerce_good_values = [({'1', 2}, {1, 2})]
        self.coerce_bad_values = [{'foo'}]
        self.to_primitive_values = [({1}, tuple([1]))]
        self.from_primitive_values = [(tuple([1]), {1})]

    def test_stringify(self):
        self.assertEqual('set([1,2])', self.field.stringify({1, 2}))

    def test_repr(self):
        self.assertEqual(
            "Set(default=<class 'oslo_versionedobjects.fields."
            "UnspecifiedDefault'>,nullable=False)",
            repr(self.field),
        )
        self.assertEqual(
            "Set(default=set([]),nullable=False)",
            repr(fields.SetOfIntegersField(default=set())),
        )
        self.assertEqual(
            "Set(default=set([1,a]),nullable=False)",
            repr(fields.SetOfIntegersField(default={1, 'a'})),  # type: ignore
        )


class TestListOfSetsOfIntegers(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.ListOfSetsOfIntegersField()
        self.coerce_good_values = [([{'1', 2}, {3, '4'}], [{1, 2}, {3, 4}])]
        self.coerce_bad_values = [[{'foo'}]]
        self.to_primitive_values = [([{1}], [tuple([1])])]
        self.from_primitive_values = [([tuple([1])], [{1}])]

    def test_stringify(self):
        self.assertEqual('[set([1,2])]', self.field.stringify([{1, 2}]))


class TestListOfIntegers(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.ListOfIntegersField()
        self.coerce_good_values = [(['1', 2], [1, 2]), ([1, 2], [1, 2])]
        self.coerce_bad_values = [['foo']]
        self.to_primitive_values = [([1], [1])]
        self.from_primitive_values = [([1], [1])]

    def test_stringify(self):
        self.assertEqual('[[1, 2]]', self.field.stringify([[1, 2]]))


class TestListOfUUIDField(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.ListOfUUIDField()
        self.uuid1 = '6b2097ea-d0e3-44dd-b131-95472b3ea8fd'
        self.uuid2 = '478c193d-2533-4e71-ab2b-c7683f67d7f9'
        self.coerce_good_values = [
            ([self.uuid1, self.uuid2], [self.uuid1, self.uuid2])
        ]
        # coerce_bad_values is intentionally ignored since the UUID field
        # allows non-UUID values for now. See TestUUIDField for examples.
        self.to_primitive_values = [([self.uuid1], [self.uuid1])]
        self.from_primitive_values = [([self.uuid1], [self.uuid1])]

    def test_stringify(self):
        self.assertEqual(
            f'[{self.uuid1},{self.uuid2}]',
            self.field.stringify([self.uuid1, self.uuid2]),
        )


class TestLocalMethods(test.TestCase):
    @mock.patch.object(obj_base.LOG, 'exception')
    def test__make_class_properties_setter_value_error(self, mock_log):
        @obj_base.VersionedObjectRegistry.register
        class AnObject(obj_base.VersionedObject):
            fields = {
                'intfield': fields.IntegerField(),
            }

        self.assertRaises(ValueError, AnObject, intfield='badvalue')
        self.assertFalse(mock_log.called)

    @mock.patch.object(obj_base.LOG, 'exception')
    def test__make_class_properties_setter_setattr_fails(self, mock_log):
        @obj_base.VersionedObjectRegistry.register
        class AnObject(obj_base.VersionedObject):
            fields = {
                'intfield': fields.IntegerField(),
            }

        # We want the setattr() call in _make_class_properties.setter() to
        # raise an exception
        with mock.patch.object(obj_base, '_get_attrname') as mock_attr:
            mock_attr.return_value = '__class__'
            self.assertRaises(TypeError, AnObject, intfield=2)
            mock_attr.assert_called_once_with('intfield')
            mock_log.assert_called_once_with(
                mock.ANY, {'attr': 'AnObject.intfield'}
            )


class TestObject(TestField):
    def setUp(self):
        super().setUp()

        @obj_base.VersionedObjectRegistry.register
        class TestableObject(obj_base.VersionedObject):
            fields = {
                'uuid': fields.StringField(),
            }

            def __eq__(self, value):
                # NOTE(danms): Be rather lax about this equality thing to
                # satisfy the assertEqual() in test_from_primitive(). We
                # just want to make sure the right type of object is re-created
                return value.__class__.__name__ == TestableObject.__name__

        class OtherTestableObject(obj_base.VersionedObject):
            pass

        test_inst = TestableObject()
        self._test_cls = TestableObject
        self.field = fields.Field(fields.Object('TestableObject'))
        self.coerce_good_values = [(test_inst, test_inst)]
        self.coerce_bad_values = [OtherTestableObject(), 1, 'foo']
        self.to_primitive_values = [(test_inst, test_inst.obj_to_primitive())]
        self.from_primitive_values = [
            (test_inst.obj_to_primitive(), test_inst),
            (test_inst, test_inst),
        ]

    def test_stringify(self):
        obj = self._test_cls(uuid='fake-uuid')
        self.assertEqual(
            'TestableObject(fake-uuid)', self.field.stringify(obj)
        )

    def test_from_primitive(self):
        @obj_base.VersionedObjectRegistry.register
        class TestFakeObject(obj_base.VersionedObject):
            OBJ_PROJECT_NAMESPACE = 'fake-project'

        @obj_base.VersionedObjectRegistry.register
        class TestBar(TestFakeObject, obj_base.ComparableVersionedObject):
            fields = {
                'name': fields.StringField(),
            }

        @obj_base.VersionedObjectRegistry.register
        class TestFoo(TestFakeObject, obj_base.ComparableVersionedObject):
            fields = {
                'name': fields.StringField(),
                'bar': fields.ObjectField('TestBar', nullable=True),
            }

        bar = TestBar(name='bar')
        foo = TestFoo(name='foo', bar=bar)
        from_primitive_values = [(foo.obj_to_primitive(), foo), (foo, foo)]

        for prim_val, out_val in from_primitive_values:
            self.assertEqual(
                out_val, self.field.from_primitive(foo, 'attr', prim_val)
            )

    def test_inheritance(self):
        # We need a whole lot of classes in a hierarchy to
        # test subclass recognition for the Object field
        class TestAnimal(obj_base.VersionedObject):
            pass

        class TestMammal(TestAnimal):
            pass

        class TestReptile(TestAnimal):
            pass

        # We'll use this to create a diamond in the
        # class hierarchy
        class TestPet(TestAnimal):
            pass

        # Non-versioned object mixin
        class TestScary:
            pass

        class TestCrocodile(TestReptile, TestPet, TestScary):
            pass

        class TestPig(TestMammal):
            pass

        class TestDog(TestMammal, TestPet):
            pass

        # Some fictional animals
        wolfy = TestDog()  # Terminator-2
        ticktock = TestCrocodile()  # Peter Pan
        babe = TestPig()  # Babe

        # The various classes
        animals = fields.Object('TestAnimal', subclasses=True)
        mammals = fields.Object('TestMammal', subclasses=True)
        reptiles = fields.Object('TestReptile', subclasses=True)
        pets = fields.Object('TestPet', subclasses=True)
        pigs = fields.Object('TestPig', subclasses=True)
        dogs = fields.Object('TestDog', subclasses=True)
        crocs = fields.Object('TestCrocodile', subclasses=True)

        self.assertEqual(
            [
                "TestDog",
                "TestMammal",
                "TestPet",
                "TestAnimal",
                "VersionedObject",
            ],
            fields.Object._get_all_obj_names(wolfy),
        )

        self.assertEqual(
            [
                "TestCrocodile",
                "TestReptile",
                "TestPet",
                "TestAnimal",
                "VersionedObject",
            ],
            fields.Object._get_all_obj_names(ticktock),
        )

        self.assertEqual(
            ["TestPig", "TestMammal", "TestAnimal", "VersionedObject"],
            fields.Object._get_all_obj_names(babe),
        )

        # When stringifying we should see the subclass object name
        # not the base class object name
        self.assertEqual("TestDog", animals.stringify(wolfy))
        self.assertEqual("TestCrocodile", animals.stringify(ticktock))
        self.assertEqual("TestPig", animals.stringify(babe))

        # Everything is an animal
        self.assertEqual(wolfy, animals.coerce(None, "animal", wolfy))
        self.assertEqual(ticktock, animals.coerce(None, "animal", ticktock))
        self.assertEqual(babe, animals.coerce(None, "animal", babe))

        # crocodiles are not mammals
        self.assertEqual(wolfy, mammals.coerce(None, "animal", wolfy))
        self.assertRaises(ValueError, mammals.coerce, None, "animal", ticktock)
        self.assertEqual(babe, mammals.coerce(None, "animal", babe))

        # dogs and pigs are not reptiles
        self.assertRaises(ValueError, reptiles.coerce, None, "animal", wolfy)
        self.assertEqual(ticktock, reptiles.coerce(None, "animal", ticktock))
        self.assertRaises(ValueError, reptiles.coerce, None, "animal", babe)

        # pigs are not pets, but crocodiles (!) & dogs are
        self.assertEqual(wolfy, pets.coerce(None, "animal", wolfy))
        self.assertEqual(ticktock, pets.coerce(None, "animal", ticktock))
        self.assertRaises(ValueError, pets.coerce, None, "animal", babe)

        # Only dogs are dogs
        self.assertEqual(wolfy, dogs.coerce(None, "animal", wolfy))
        self.assertRaises(ValueError, dogs.coerce, None, "animal", ticktock)
        self.assertRaises(ValueError, dogs.coerce, None, "animal", babe)

        # Only crocs are crocs
        self.assertRaises(ValueError, crocs.coerce, None, "animal", wolfy)
        self.assertEqual(ticktock, crocs.coerce(None, "animal", ticktock))
        self.assertRaises(ValueError, crocs.coerce, None, "animal", babe)

        # Only pigs are pigs
        self.assertRaises(ValueError, pigs.coerce, None, "animal", ticktock)
        self.assertRaises(ValueError, pigs.coerce, None, "animal", wolfy)
        self.assertEqual(babe, pigs.coerce(None, "animal", babe))

    def test_coerce_bad_value_primitive_type(self):
        # Tests that the ValueError has the primitive type in it's message.
        ex = self.assertRaises(
            ValueError, self.field.coerce, 'obj', 'attr', [{}]
        )
        self.assertEqual(
            'An object of type TestableObject is required '
            'in field attr, not a list',
            str(ex),
        )

    def test_get_schema(self):
        self.assertEqual(
            {
                'properties': {
                    'versioned_object.changes': {
                        'items': {'type': 'string'},
                        'type': 'array',
                    },
                    'versioned_object.data': {
                        'description': 'fields of TestableObject',
                        'properties': {
                            'uuid': {'readonly': False, 'type': ['string']}
                        },
                        'required': ['uuid'],
                        'type': 'object',
                    },
                    'versioned_object.name': {'type': 'string'},
                    'versioned_object.namespace': {'type': 'string'},
                    'versioned_object.version': {'type': 'string'},
                },
                'readonly': False,
                'required': [
                    'versioned_object.namespace',
                    'versioned_object.name',
                    'versioned_object.version',
                    'versioned_object.data',
                ],
                'type': ['object'],
            },
            self.field.get_schema(),
        )


class TestIPAddress(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.IPAddressField()
        self.coerce_good_values = [
            ('1.2.3.4', netaddr.IPAddress('1.2.3.4')),
            ('::1', netaddr.IPAddress('::1')),
            (netaddr.IPAddress('::1'), netaddr.IPAddress('::1')),
        ]
        self.coerce_bad_values = ['1-2', 'foo']
        self.to_primitive_values = [
            (netaddr.IPAddress('1.2.3.4'), '1.2.3.4'),
            (netaddr.IPAddress('::1'), '::1'),
        ]
        self.from_primitive_values = [
            ('1.2.3.4', netaddr.IPAddress('1.2.3.4')),
            ('::1', netaddr.IPAddress('::1')),
        ]


class TestIPAddressV4(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.IPV4AddressField()
        self.coerce_good_values = [
            ('1.2.3.4', netaddr.IPAddress('1.2.3.4')),
            (netaddr.IPAddress('1.2.3.4'), netaddr.IPAddress('1.2.3.4')),
        ]
        self.coerce_bad_values = ['1-2', 'foo', '::1']
        self.to_primitive_values = [(netaddr.IPAddress('1.2.3.4'), '1.2.3.4')]
        self.from_primitive_values = [
            ('1.2.3.4', netaddr.IPAddress('1.2.3.4'))
        ]

    def test_get_schema(self):
        self.assertEqual(
            {'type': ['string'], 'readonly': False, 'format': 'ipv4'},
            self.field.get_schema(),
        )


class TestIPAddressV6(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.IPV6AddressField()
        self.coerce_good_values = [
            ('::1', netaddr.IPAddress('::1')),
            (netaddr.IPAddress('::1'), netaddr.IPAddress('::1')),
        ]
        self.coerce_bad_values = ['1.2', 'foo', '1.2.3.4']
        self.to_primitive_values = [(netaddr.IPAddress('::1'), '::1')]
        self.from_primitive_values = [('::1', netaddr.IPAddress('::1'))]

    def test_get_schema(self):
        self.assertEqual(
            {'type': ['string'], 'readonly': False, 'format': 'ipv6'},
            self.field.get_schema(),
        )


class TestIPV4AndV6Address(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.IPV4AndV6Address()
        self.coerce_good_values = [
            ('::1', netaddr.IPAddress('::1')),
            (netaddr.IPAddress('::1'), netaddr.IPAddress('::1')),
            ('1.2.3.4', netaddr.IPAddress('1.2.3.4')),
            (netaddr.IPAddress('1.2.3.4'), netaddr.IPAddress('1.2.3.4')),
        ]
        self.coerce_bad_values = ['1-2', 'foo']
        self.to_primitive_values = [
            (netaddr.IPAddress('::1'), '::1'),
            (netaddr.IPAddress('1.2.3.4'), '1.2.3.4'),
        ]
        self.from_primitive_values = [
            ('::1', netaddr.IPAddress('::1')),
            ('1.2.3.4', netaddr.IPAddress('1.2.3.4')),
        ]

    def test_get_schema(self):
        self.assertEqual(
            {
                'oneOf': [
                    {'format': 'ipv4', 'type': ['string']},
                    {'format': 'ipv6', 'type': ['string']},
                ]
            },
            self.field.get_schema(),
        )


class TestIPNetwork(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.IPNetworkField()
        self.coerce_good_values = [
            ('::1/0', netaddr.IPNetwork('::1/0')),
            ('1.2.3.4/24', netaddr.IPNetwork('1.2.3.4/24')),
            (netaddr.IPNetwork('::1/32'), netaddr.IPNetwork('::1/32')),
        ]
        self.coerce_bad_values = ['foo']
        self.to_primitive_values = [(netaddr.IPNetwork('::1/0'), '::1/0')]
        self.from_primitive_values = [('::1/0', netaddr.IPNetwork('::1/0'))]


class TestIPV4Network(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.IPV4NetworkField()
        self.coerce_good_values = [
            ('1.2.3.4/24', netaddr.IPNetwork('1.2.3.4/24'))
        ]
        self.coerce_bad_values = ['foo', '::1/32']
        self.to_primitive_values = [
            (netaddr.IPNetwork('1.2.3.4/24'), '1.2.3.4/24')
        ]
        self.from_primitive_values = [
            ('1.2.3.4/24', netaddr.IPNetwork('1.2.3.4/24'))
        ]

    def test_get_schema(self):
        schema = self.field.get_schema()
        self.assertEqual(['string'], schema['type'])
        self.assertEqual(False, schema['readonly'])
        pattern = schema['pattern']
        for _, valid_val in self.coerce_good_values:
            self.assertRegex(str(valid_val), pattern)
        invalid_vals = [x for x in self.coerce_bad_values]
        for invalid_val in invalid_vals:
            self.assertNotRegex(str(invalid_val), pattern)


class TestIPV6Network(TestField):
    def setUp(self):
        super().setUp()
        self.field = fields.IPV6NetworkField()
        self.coerce_good_values = [
            ('::1/0', netaddr.IPNetwork('::1/0')),
            (netaddr.IPNetwork('::1/32'), netaddr.IPNetwork('::1/32')),
        ]
        self.coerce_bad_values = ['foo', '1.2.3.4/24']
        self.to_primitive_values = [(netaddr.IPNetwork('::1/0'), '::1/0')]
        self.from_primitive_values = [('::1/0', netaddr.IPNetwork('::1/0'))]

    def test_get_schema(self):
        schema = self.field.get_schema()
        self.assertEqual(['string'], schema['type'])
        self.assertEqual(False, schema['readonly'])
        pattern = schema['pattern']
        for _, valid_val in self.coerce_good_values:
            self.assertRegex(str(valid_val), pattern)
        invalid_vals = [x for x in self.coerce_bad_values]
        for invalid_val in invalid_vals:
            self.assertNotRegex(str(invalid_val), pattern)


class FakeCounter:
    def __init__(self):
        self.n = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.n <= 4:
            self.n += 1
            return self.n
        else:
            raise StopIteration


class TestListTypes(test.TestCase):
    def test_regular_list(self):
        fields.List(fields.Integer()).coerce(None, '', [1, 2])

    def test_non_iterable(self):
        self.assertRaises(
            ValueError, fields.List(fields.Integer()).coerce, None, '', 2
        )

    def test_string_iterable(self):
        self.assertRaises(
            ValueError,
            fields.List(fields.Integer()).coerce,
            None,
            None,
            'hello',
        )

    def test_mapping_iterable(self):
        self.assertRaises(
            ValueError,
            fields.List(fields.Integer()).coerce,
            None,
            None,
            {'a': 1, 'b': 2},
        )

    def test_iter_class(self):
        fields.List(fields.Integer()).coerce(None, '', FakeCounter())



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""VersionedObjects base exception handling.

Includes decorator for re-raising VersionedObjects-type exceptions.

SHOULD include dedicated exception logging.

"""

from collections.abc import Callable
import functools
import inspect
import logging
from typing import Any, Concatenate, ParamSpec, TypeVar

from oslo_config import cfg
from oslo_utils import excutils

from oslo_versionedobjects._i18n import _

LOG: logging.Logger = logging.getLogger(__name__)

exc_log_opts: list[cfg.Opt] = [
    cfg.BoolOpt(
        'fatal_exception_format_errors',
        default=False,
        help='Make exception message format errors fatal',
    ),
]

CONF: cfg.ConfigOpts = cfg.CONF
CONF.register_opts(exc_log_opts, group='oslo_versionedobjects')


def _cleanse_dict(original: dict[str, Any]) -> dict[str, Any]:
    """Strip all admin_password, new_pass, rescue_pass keys from a dict."""
    return {k: v for k, v in original.items() if "_pass" not in k}


_F = TypeVar('_F', bound=Callable[..., Any])

P = ParamSpec('P')
R = TypeVar('R')


def wrap_exception(
    notifier: Any = None, get_notifier: Callable[[], Any] | None = None
) -> Callable[
    [Callable[Concatenate[Any, P], R]],
    Callable[Concatenate[Any, P], R | None],
]:
    """Catch all exceptions in wrapped method

    This decorator wraps a method to catch any exceptions that may
    get thrown. It also optionally sends the exception to the notification
    system.
    """

    def inner(
        f: Callable[Concatenate[Any, P], R],
    ) -> Callable[Concatenate[Any, P], R | None]:
        def wrapped(
            self: Any, /, context: Any, *args: P.args, **kw: P.kwargs
        ) -> R | None:
            # Don't store self or context in the payload, it now seems to
            # contain confidential information.
            try:
                return f(self, context, *args, **kw)
            except Exception as e:
                with excutils.save_and_reraise_exception():
                    if notifier or get_notifier:
                        payload: dict[str, Any] = dict(exception=e)
                        call_dict = inspect.getcallargs(
                            f, self, context, *args, **kw
                        )
                        cleansed = _cleanse_dict(call_dict)
                        payload.update({'args': cleansed})

                        # If f has multiple decorators, they must use
                        # functools.wraps to ensure the name is
                        # propagated.
                        event_type = f.__name__

                        if notifier:
                            actual_notifier = notifier
                        else:
                            # get_notifier must be set since we're in
                            # the 'if notifier or get_notifier' block
                            actual_notifier = get_notifier()  # type: ignore[misc]
                        actual_notifier.error(context, event_type, payload)
                return None

        return functools.wraps(f)(wrapped)  # type: ignore[return-value]

    return inner


class VersionedObjectsException(Exception):
    """Base VersionedObjects Exception

    To correctly use this class, inherit from it and define
    a 'msg_fmt' property. That msg_fmt will get printf'd
    with the keyword arguments provided to the constructor.
    """

    msg_fmt = _("An unknown exception occurred.")

    def __init__(self, message: str | None = None, **kwargs: Any) -> None:
        self.kwargs: dict[str, Any] = kwargs

        if 'code' not in self.kwargs:
            try:
                self.kwargs['code'] = 500
            except AttributeError:
                pass

        if not message:
            try:
                message = self.msg_fmt % kwargs
            except Exception:
                # kwargs doesn't match a variable in the message
                # log the issue and the kwargs
                LOG.exception('Exception in string format operation')
                for name, value in kwargs.items():
                    LOG.error(f"{name}: {value}")  # noqa

                if CONF.oslo_versionedobjects.fatal_exception_format_errors:
                    raise
                else:
                    # at least get the core message out if something happened
                    message = self.msg_fmt

        super().__init__(message)

    def format_message(self) -> str:
        # NOTE(mrodden): use the first argument to the python Exception object
        # which should be our full VersionedObjectsException message,
        # (see __init__)
        return str(self.args[0])


class ObjectActionError(VersionedObjectsException):
    msg_fmt = _('Object action %(action)s failed because: %(reason)s')


class ObjectFieldInvalid(VersionedObjectsException):
    msg_fmt = _('Field %(field)s of %(objname)s is not an instance of Field')


class OrphanedObjectError(VersionedObjectsException):
    msg_fmt = _('Cannot call %(method)s on orphaned %(objtype)s object')


class IncompatibleObjectVersion(VersionedObjectsException):
    msg_fmt = _(
        'Version %(objver)s of %(objname)s is not supported, '
        'supported version is %(supported)s'
    )


class ReadOnlyFieldError(VersionedObjectsException):
    msg_fmt = _('Cannot modify readonly field %(field)s')


class UnsupportedObjectError(VersionedObjectsException):
    msg_fmt = _('Unsupported object type %(objtype)s')


class EnumRequiresValidValuesError(VersionedObjectsException):
    msg_fmt = _('Enum fields require a list of valid_values')


class EnumValidValuesInvalidError(VersionedObjectsException):
    msg_fmt = _('Enum valid values are not valid')


class EnumFieldInvalid(VersionedObjectsException):
    msg_fmt = _('%(typename)s in %(fieldname)s is not an instance of Enum')


class EnumFieldUnset(VersionedObjectsException):
    msg_fmt = _('%(fieldname)s missing field type')


class InvalidTargetVersion(VersionedObjectsException):
    msg_fmt = _('Invalid target version %(version)s')


class TargetBeforeSubobjectExistedException(VersionedObjectsException):
    msg_fmt = _("No subobject existed at version %(target_version)s")


class UnregisteredSubobject(VersionedObjectsException):
    msg_fmt = _(
        "%(child_objname)s is referenced by %(parent_objname)s but "
        "is not registered"
    )
